"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Circuit, Stabilizer
from loom.eka.utilities import Direction
from loom.interpreter import InterpretationStep

from .y_wall_out_utilities import find_relative_diagonal_direction
from ..move_block import (
    DetailedSchedule,
    generate_syndrome_measurement_circuit_and_cbits,
    generate_and_append_block_syndromes_and_detectors,
)
from ..utilities import FourBodySchedule
from ...code_factory import RotatedSurfaceCode


def y_wall_out_initial_syndrome_measurement(
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    same_timeslice: bool,
    debug_mode: bool,  # pylint: disable=unused-argument
) -> InterpretationStep:
    """
    Generates and appends the initial syndrome measurement circuit for the
    Y-wall-out applicator. This syndrome measurement is performed on the
    initial block before the y-wall-out operation and is crafted such that the whole
    operation remains fault-tolerant.


    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step.
    block : RotatedSurfaceCode
        The block to which the operation will be applied.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block.
        (Currently unused since no new Block is created here.)

    Returns
    -------
    InterpretationStep
        The updated interpretation step with the initial syndrome measurement
    """

    # Initialization circuit
    init_block_syndrome_measurement_reset_circuit = Circuit(
        name="Initialization of syndrome measurement ancilla",
        circuit=[
            [
                Circuit(
                    f"reset_{'0' if stab.pauli_type == 'Z' else '+'}",
                    channels=[
                        interpretation_step.get_channel_MUT(stab.ancilla_qubits[0])
                    ],
                )
                for stab in block.stabilizers
            ]
        ],
    )

    # CNOTs circuit
    init_block_syndrome_measurement_cnot_circuit = (
        get_initial_block_syndrome_measurement_cnots_circuit(block, interpretation_step)
    )

    # Measurements and generation of cbits
    # (their order matches the stabilizer order in the block)
    init_block_syndrome_measurement_measure_circuit, init_cbits = (
        generate_syndrome_measurement_circuit_and_cbits(
            interpretation_step,
            block,
        )
    )

    # Assemble and append circuit
    interpretation_step.append_circuit_MUT(
        Circuit(
            name="Y wall out - Initial block syndrome measurement",
            circuit=(
                init_block_syndrome_measurement_reset_circuit,
                init_block_syndrome_measurement_cnot_circuit,
                init_block_syndrome_measurement_measure_circuit,
            ),
        ),
        same_timeslice=same_timeslice,
    )

    # Generate and append syndromes and detectors
    generate_and_append_block_syndromes_and_detectors(
        interpretation_step, block, init_cbits
    )

    return interpretation_step


def get_initial_block_syndrome_measurement_cnots_circuit(
    # pylint: disable=too-many-locals
    block: RotatedSurfaceCode,
    interpretation_step: InterpretationStep,
) -> Circuit:
    """
    Generates fault-tolerant CNOT scheduling for syndrome measurement of
    the initial block.
    The block is divided into regions as explained below.

    Examples: a 7x14 block with three lines corresponding to three ways
    for dividing the block::

                3 * * * * * 4
                *         . *
                * .     .   *
                *   . .     *
                *   . .     *
                * .     .   *
                o         . *
                * .         2
                *   .       *
                *     .     *
                *       .   *
                *         . *
                *           *
                1 * * * * * *

    - Region 1: draw a diagonal line from corner 2 (middle-edge corner) \
    towards corner 3 (angle corner), note that the line doesn't pass through \
    corner 3). The line split the data qubits into two regions. \
    Region 1 is defined as the region containing corner 4 (short-end corner).
    - Region 2: draw a diagonal line from qubit o to corner 4. Qubit o is the qubit \
    shifted by 1 from the opposite of corner 2 towards corner 3. Region 2 is defined \
     as the region containing corner 3.
    - Region 3: draw a diagonal line from qubit o towards the non-topological \
    geometric corner. Region 3 is defined as the region containing corner 1.
    - Region 4: qubits not contained in region 1, 2, or 3.

    The scheduling algorithm is the following.

    - Z stabilizers have the same schedule across the block. The schedule is such that \
    Z errors propagating perpendicular to the edge connecting corners 3 and 4.
    - X stabilizers in region 1, 2, and 3 are scheduled opposite to the schedule of \
    Z stabilizers.
    - X stabilizers in region 4 are scheduled opposite to the schedule of X \
    stabilizers in region 1, 2, and 3.

    Parameters
    ----------
    block : RotatedSurfaceCode
        The block to which the operation will be applied.
    interpretation_step : InterpretationStep
        The interpretation step for the circuit generation.

    Returns
    -------
    Circuit
        The generated CNOT circuit for the syndrome measurement of the initial block.

    """
    config, pivot_corners = block.config_and_pivot_corners
    if config != 4:
        raise ValueError(
            "The initial block of Y-wall-out must have type-4 corner configuration."
        )

    long_end_corner = pivot_corners[0]
    top_left_corner = block.upper_left_qubit
    is_horizontal = block.is_horizontal

    # The starting qubit direction is the direction of the long-end corner
    starting_diag_direction = find_relative_diagonal_direction(
        top_left_corner, long_end_corner
    )

    # Schedules
    # Z schedule is determined based on the orientation of the short edge
    z_schedule = FourBodySchedule.Z if is_horizontal else FourBodySchedule.N
    region_123_x_schedule = z_schedule.opposite_schedule()
    region_4_x_schedule = region_123_x_schedule.opposite_schedule()

    # Determine cnot sequence, require 5 time steps
    cnots = [[] for _ in range(5)]
    for stab in block.stabilizers:
        pauli_type = stab.pauli_type
        (
            is_in_region_4,
            is_on_boundary_12_and_in_region_4,
            is_closer_long_end_corner,
            boundary_direction,
        ) = find_stabilizer_position_in_initial_block(block, stab, pivot_corners)
        if is_in_region_4:
            schedule = region_4_x_schedule if pauli_type == "X" else z_schedule
        else:
            schedule = region_123_x_schedule if pauli_type == "X" else z_schedule

        detailed_schedule = DetailedSchedule.from_schedule_and_direction(
            schedule, starting_diag_direction
        )

        if len(stab.data_qubits) == 4:
            data_qubits = detailed_schedule.get_stabilizer_qubits(stab)
        else:
            data_qubits = detailed_schedule.get_stabilizer_qubits(
                stab, boundary_direction
            )

        for i, data_qubit in enumerate(data_qubits):
            if data_qubit is None:
                continue
            if pauli_type == "Z":
                # If the stabilizer is Z, cnot from data qubit to ancilla
                cnot_pair = (data_qubit, stab.ancilla_qubits[0])
            elif pauli_type == "X":
                # If the stabilizer is X, cnot from ancilla to data qubit
                cnot_pair = (stab.ancilla_qubits[0], data_qubit)
            else:
                raise ValueError("Unknown stabilizer type.")

            # Determine the cnot layer
            layer = (
                i + 1
                if (
                    # long end corner half
                    is_closer_long_end_corner
                    # bulk of region 4
                    or (is_in_region_4 and not is_on_boundary_12_and_in_region_4)
                    # boundary of region 4
                    or (
                        is_in_region_4
                        and is_on_boundary_12_and_in_region_4
                        and i in [2, 3]
                    )
                )
                else i
            )
            cnots[layer].append(cnot_pair)

    # Generate the circuit containing the cnots
    cnots_circuit = Circuit(
        name="Initial block syndrome measurement CNOT circuit",
        circuit=[
            [
                Circuit(
                    "cx",
                    channels=[
                        interpretation_step.get_channel_MUT(q) for q in qubit_pair
                    ],
                )
                for qubit_pair in cnot_slice
            ]
            for cnot_slice in cnots
        ],
    )

    return cnots_circuit


def find_stabilizer_position_in_initial_block(
    block: RotatedSurfaceCode,
    stab: Stabilizer,
    pivot_corners: tuple[
        tuple[int, ...], tuple[int, ...], tuple[int, ...], tuple[int, ...]
    ],
) -> tuple[bool, bool, bool, Direction | None]:
    # pylint: disable=too-many-locals
    """
    Find stabilizer position with respect to the regions of the initial block

    Parameters
    ----------
    block: RotatedSurfaceCode
        The initial block
    stab: Stabilizer
        The input stabilizer
    pivot_corners:
        tuple[tuple[int, ...], tuple[int, ...], tuple[int, ...], tuple[int, ...]]
        Four topological corners

    Returns
    -------
    tuple[bool, bool, bool, Direction | None]
        True if the stabilizer is inside region 4
        True if the stabilizer is inside region 4 and next to region 1 or 2
        True if the stabilizer is in the half of the block containing corner 1 and 2
        Direction of the boundary if the input stabilizer is a weight-2 stabilizer

    """

    long_end_corner, middle_corner = pivot_corners[0], pivot_corners[1]
    is_horizontal = block.is_horizontal
    long_edge_idx = 0 if is_horizontal else 1  # coordinate index along the long side
    short_edge_idx = 1 if is_horizontal else 0  # coordinate index along the short side

    def is_in_region_1(q) -> bool:
        """Determine if the input qubit is inside region 1
        (including the boundaries of the region). This is done by checking
        whether the distance along the long side of the block between the input
        qubit and the middle-edge corner is at least the distance along the
        short side between the same two qubits.
        """
        # Compare middle_corner and long_end_corner along the short edge to determine
        # if the qubit is on the left or right of middle_corner
        # Distance is calculated with respect to middle_corner
        dshort = (
            middle_corner[short_edge_idx] - q[short_edge_idx]
            if middle_corner[short_edge_idx] > long_end_corner[short_edge_idx]
            else q[short_edge_idx] - middle_corner[short_edge_idx]
        )
        # Compare middle_corner and long_end_corner along the long edge to determine
        # if the qubit is on the top or bottom of middle_corner
        # Distance is calculated with respect to middle_corner
        dlong = (
            middle_corner[long_edge_idx] - q[long_edge_idx]
            if middle_corner[long_edge_idx] < long_end_corner[long_edge_idx]
            else q[long_edge_idx] - middle_corner[long_edge_idx]
        )
        return 0 <= dshort <= dlong and dlong >= 0

    def is_in_region_2(q) -> bool:
        """Determine if the input qubit is inside region 2
        (including the boundaries of the region). This is done by checking
        whether the distance along the long side of the block between the input
        qubit and the qubit o is at least the distance along the
        short side between the same two qubits.
        """
        # Compare middle_corner and long_end_corner along the short edge to determine
        # if the qubit is on the left or right of middle_corner.
        # Distance is calculated with respect to long_end_corner because it has the
        # same (short-edge) coordinate as qubit o.
        dshort = (
            q[short_edge_idx] - long_end_corner[short_edge_idx]
            if middle_corner[short_edge_idx] > long_end_corner[short_edge_idx]
            else long_end_corner[short_edge_idx] - q[short_edge_idx]
        )
        # Compare middle_corner and long_end_corner along the long edge to determine
        # if the qubit is on the top or bottom of middle_corner.
        # Distance is calculated with respect to middle_corner because the (long-edge)
        # coordinate of qubit o can be deduced from it.
        dlong = (
            middle_corner[long_edge_idx] - q[long_edge_idx] - 1
            if middle_corner[long_edge_idx] < long_end_corner[long_edge_idx]
            else q[long_edge_idx] - middle_corner[long_edge_idx] - 1
        )
        return 0 <= dshort <= dlong and dlong >= 0

    def is_in_region_3(q) -> bool:
        """Determine if the input qubit is inside region 2
        (including the boundaries of the region). This is done by checking
        whether the distance along the long side of the block between the input
        qubit and the qubit o is at least the distance along the
        short side between the same two qubits.
        """
        # Compare middle_corner and long_end_corner along the short edge to determine
        # if the qubit is on the left or right of middle_corner.
        # Distance is calculated with respect to long_end_corner because it has the
        # same (short-edge) coordinate as qubit o.
        dshort = (
            q[short_edge_idx] - long_end_corner[short_edge_idx]
            if middle_corner[short_edge_idx] > long_end_corner[short_edge_idx]
            else long_end_corner[short_edge_idx] - q[short_edge_idx]
        )
        # Compare middle_corner and long_end_corner along the long edge to determine
        # if the qubit is on the top or bottom of middle_corner.
        # Distance is calculated with respect to middle_corner because the (long-edge)
        # coordinate of qubit o can be deduced from it.
        # Note that region 3 is on the opposite side (along the long edge) of region 2
        # with respect to qubit o.
        dlong = (
            q[long_edge_idx] - middle_corner[long_edge_idx]
            if middle_corner[long_edge_idx] < long_end_corner[long_edge_idx]
            else middle_corner[long_edge_idx] - q[long_edge_idx]
        )
        return 0 <= dshort <= dlong and dlong >= 0

    dq_in_region_1 = [dqubit for dqubit in stab.data_qubits if is_in_region_1(dqubit)]
    dq_in_region_2 = [dqubit for dqubit in stab.data_qubits if is_in_region_2(dqubit)]
    dq_in_region_3 = [dqubit for dqubit in stab.data_qubits if is_in_region_3(dqubit)]

    is_stab_in_region_1 = len(dq_in_region_1) == len(stab.data_qubits)
    is_stab_in_region_2 = len(dq_in_region_2) == len(stab.data_qubits)
    is_stab_in_region_3 = len(dq_in_region_3) == len(stab.data_qubits)
    is_stab_in_region_4 = not (
        is_stab_in_region_1 or is_stab_in_region_2 or is_stab_in_region_3
    )

    is_stab_on_boundary_region_1 = (len(stab.data_qubits) - len(dq_in_region_1)) == 1
    is_stab_on_boundary_region_2 = (len(stab.data_qubits) - len(dq_in_region_2)) == 1
    is_stab_in_region_4_and_on_boundary_region_12 = is_stab_in_region_4 and (
        is_stab_on_boundary_region_1 or is_stab_on_boundary_region_2
    )

    # Find if the stabilizer is near the long-end corner
    top_left_qubit = max(stab.data_qubits, key=lambda x: -x[0] - x[1])
    bottom_right_qubit = max(stab.data_qubits, key=lambda x: x[0] + x[1])
    compare_qubit = (
        top_left_qubit
        if middle_corner[long_edge_idx] < long_end_corner[long_edge_idx]
        else bottom_right_qubit
    )
    is_closer_long_end_corner = (
        middle_corner[long_edge_idx]
        <= compare_qubit[long_edge_idx]
        <= long_end_corner[long_edge_idx]
        if long_end_corner[long_edge_idx] > middle_corner[long_edge_idx]
        else long_end_corner[long_edge_idx]
        <= compare_qubit[long_edge_idx]
        <= middle_corner[long_edge_idx]
    )

    # Find the boundary direction of weight-2 stabilizer
    boundary = None
    weight = len(stab.pauli)
    if weight == 2:
        qubit_directions = [
            set(
                direction
                for direction in Direction
                if dqubit in block.boundary_qubits(direction)
            )
            for dqubit in stab.data_qubits
        ]
        boundary = list(qubit_directions[0].intersection(qubit_directions[1]))[0]

    return (
        is_stab_in_region_4,
        is_stab_in_region_4_and_on_boundary_region_12,
        is_closer_long_end_corner,
        boundary,
    )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=duplicate-code
from __future__ import annotations
from uuid import uuid4
from pydantic.dataclasses import dataclass

from loom.eka import (
    Block,
    Lattice,
    LatticeType,
    PauliOperator,
    Stabilizer,
)

from loom.eka.utilities import Direction, dataclass_config


# pylint: disable=too-many-arguments,too-many-positional-arguments,too-many-branches
@dataclass(config=dataclass_config)
class RepetitionCode(Block):
    """
    A sub-class of `Block` that represents a repetition code block.
    Contains methods to create a repetition code block.
    """

    @classmethod
    def create(  # pylint: disable=too-many-arguments,too-many-positional-arguments,too-many-branches
        cls,
        d: int,
        check_type: str,
        lattice: Lattice,
        unique_label: str | None = None,
        position: tuple[int, ...] = (0,),
        logical_x_operator: PauliOperator | None = None,
        logical_z_operator: PauliOperator | None = None,
        skip_validation: bool = False,
    ) -> RepetitionCode:
        """Create a `Block` object for a repetition code block. The repetition code is
        defined as a linear chain with open boundary conditions.

        Parameters
        ----------
        d : int
            Code distance and size of the chain.
        check_type : str
            Type of code stabilizers, either "X" (phase-flip) or "Z" (bit-flip).
        lattice : Lattice
            Lattice on which the block is defined. The qubit indices depend on the type
            of lattice.
        unique_label : str, optional
            Label for the block. It must be unique among all blocks in the initial CRD.
            If no label is provided, a unique label is generated automatically using the
            uuid module.
        position : tuple[int, ...], optional
            Position of the top left corner of the block on the lattice, by
            default (0,).
        logical_x_operator: PauliOperator | None, optional
            Logical X operator. For bit-flip code, if None is provided, by default
            the full chain of qubits is selected. For phase-flip code, if None is
            provided, by default the first qubit is selected.
        logical_z_operator: PauliOperator | None, optional
            Logical Z operator. For phase-flip code, if None is provided, by default
            the full chain of qubits is selected. For bit-flip code, if None is
            provided, by default the first qubit is selected.
        skip_validation : bool, optional
            Skip validation of the block object, by default False.

        Returns
        -------
        Block
            Block object for a repetition code chain
        """

        # Input validation
        # Check distance is a positive integer
        if not isinstance(d, int) or d <= 0:
            raise ValueError(f"`d` must be a positive integer. Got '{d}' instead.")

        # Check check_type is either "X" or "Z"
        if check_type not in ["X", "Z"]:
            raise ValueError(
                f"`check_type` must be either 'X' or 'Z'. Got '{check_type}' instead."
            )

        # Check lattice is linear
        if lattice.lattice_type != LatticeType.LINEAR:
            raise ValueError(
                "The creation of repetition chains is "
                "currently only supported for linear lattices. Instead "
                f"the lattice is of type {lattice.lattice_type}."
            )

        # Ensure position is a tuple of integers
        if not isinstance(position, tuple) or any(
            not isinstance(x, int) for x in position
        ):
            raise ValueError(
                f"`position` must be a tuple of integers. Got '{position}' instead."
            )

        if len(position) != lattice.n_dimensions:
            raise ValueError(
                f"`position` has length {len(position)} while length "
                f"{lattice.n_dimensions} is required to match the lattice dimension."
            )

        # Assign uuid if not label provided
        if unique_label is None:
            unique_label = str(uuid4())

        # Check logical operator lengths
        if logical_x_operator is not None and check_type == "Z":
            if len(logical_x_operator.pauli) != d:
                raise ValueError(
                    "Support of input X logical should be equal to distance"
                )

        if logical_z_operator is not None and check_type == "X":
            if len(logical_z_operator.pauli) != d:
                raise ValueError(
                    "Support of input Z logical should be equal to distance"
                )

        # Create stabilizers
        stabilizers = [
            Stabilizer(
                pauli=check_type * 2,
                data_qubits=[(i, 0), (i + 1, 0)],
                ancilla_qubits=[(i, 1)],
            )
            for i in range(d - 1)
        ]

        # Create logical operators
        if logical_x_operator is None:
            logical_x_operator = (
                PauliOperator(pauli=check_type, data_qubits=[(0, 0)])
                if check_type == "X"
                else PauliOperator(
                    pauli="X" * d, data_qubits=[(i, 0) for i in range(d)]
                )
            )
        if logical_z_operator is None:
            logical_z_operator = (
                PauliOperator(pauli=check_type, data_qubits=[(0, 0)])
                if check_type == "Z"
                else PauliOperator(
                    pauli="Z" * d, data_qubits=[(i, 0) for i in range(d)]
                )
            )
        block = cls(
            unique_label=unique_label,
            stabilizers=stabilizers,
            logical_x_operators=[logical_x_operator],
            logical_z_operators=[logical_z_operator],
            skip_validation=skip_validation,
        )
        if position == (0,):
            return block

        return block.shift(position)

    @property
    def check_type(self) -> str:
        """Extract the check type of the repetition code.

        Returns
        -------
        str
            Check type of the repetition code, either "X" or "Z"
        """
        return self.stabilizers[0].pauli[0]

    def boundary_qubits(self, direction: Direction | str) -> tuple[int, int]:
        """Return the data qubits that are part of the specified boundary.

        Parameters
        ----------
        direction : Direction | str
            Boundary (left or right) for which the data qubits should be
            returned. If a string is provided, it is converted to a Direction enum.

        Returns
        -------
        tuple[int, int]
            Data qubit in the specified boundary
        """
        # Input validation: cast direction to Direction enum if it is not already
        if not isinstance(direction, Direction):
            direction = Direction(direction)

        if direction not in [Direction.LEFT, Direction.RIGHT]:
            raise ValueError(
                f"Invalid direction '{direction}'. "
                "Only 'left' and 'right' are supported."
            )

        # NOTE: Even though the type of Block.data_qubits is tuple[int, ...], the output
        # of this function will in practice always be a tuple of two ints, as the
        # repetition code can only be applied to a linear lattice
        selector_function = max if direction is Direction.RIGHT else min

        return selector_function(self.data_qubits, key=lambda x: x[0])  # type: ignore

    def get_shifted_equivalent_logical_operator(
        self,
        new_qubit: tuple[int, int],
    ) -> tuple[PauliOperator, tuple[Stabilizer, ...]]:
        """Get a shifted version of the single qubit logical operator and the
        stabilizers required to perform the shift.

        Parameters
        ----------
        new_qubit : tuple[int, int]
            New qubit where the logical operator should be shifted.

        Returns
        -------
        tuple[PauliOperator, tuple[Stabilizer, ...]]
            Shifted logical operator and the stabilizers required to perform the shift.
        """

        # Check if the new qubit is in the data qubits
        if new_qubit not in self.data_qubits:
            raise ValueError(
                f"New logical position {new_qubit} is not part of the data qubits"
            )

        short_logical = (
            self.logical_x_operators[0]
            if self.check_type == "X"
            else self.logical_z_operators[0]
        )
        short_logical_data_qubit = short_logical.data_qubits[0]

        # Define shifted operator
        if new_qubit != short_logical_data_qubit:
            shifted_operator = PauliOperator(
                pauli=self.check_type, data_qubits=[new_qubit]
            )
        else:
            shifted_operator = short_logical

        # Define required stabilizers
        if new_qubit < short_logical_data_qubit:
            required_stabilizers = tuple(
                stab
                for stab in self.stabilizers
                if all(
                    (q[0] >= new_qubit[0] and q[0] <= short_logical_data_qubit[0])
                    for q in stab.data_qubits
                )
            )
        elif new_qubit > short_logical_data_qubit:
            required_stabilizers = tuple(
                stab
                for stab in self.stabilizers
                if all(
                    (q[0] <= new_qubit[0] and q[0] >= short_logical_data_qubit[0])
                    for q in stab.data_qubits
                )
            )
        else:
            required_stabilizers = ()

        return shifted_operator, required_stabilizers

    # Define equality method for repetition codes
    def __eq__(self, other: RepetitionCode) -> bool:
        if not isinstance(other, RepetitionCode):
            raise NotImplementedError(
                f"Cannot compare RepetitionCode with {type(other)}"
            )
        return super().__eq__(other)

    def shift(
        self, position: tuple[int], new_label: str | None = None
    ) -> RepetitionCode:
        return super().shift(position, new_label)

    def rename(self, name: str) -> RepetitionCode:
        return super().rename(name)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from typing import Union
import uuid

from pydantic import ValidationInfo
from pydantic import ConfigDict

dataclass_config = ConfigDict(frozen=True, extra="forbid")


def coordinate_length_error(
    list_of_coordinates: tuple[tuple, ...],
) -> tuple[tuple, ...]:
    """
    Check if the length of the lists in the list are consistent.
    """
    if len(list_of_coordinates) != 0:
        length = len(list_of_coordinates[0])
    for item in list_of_coordinates:
        if len(item) != length:
            raise ValueError("Length of coordinates must be consistent.")

    return list_of_coordinates


def distinct_error(value: Union[tuple, list]):
    """
    Check if the elements in a list or tuple are distinct. Throw an error otherwise.
    """
    if all(isinstance(i, Union[tuple, list]) for i in value):
        value = tuple(map(tuple, value))  # Ensure we are dealing with a tuple of tuples
    else:
        value = tuple(value)
    if len(value) != len(set(value)):
        raise ValueError(f"List of {type(value[0])} must have distinct elements.")

    return value


def empty_list_error(list_obj: list):
    """
    Check if the list is empty. Throw an error if that is the case.
    """
    if len(list_obj) == 0:
        raise ValueError("List cannot be empty.")

    return list_obj


def ensure_tuple(list_obj):
    """
    Adjusts lists to tuples to ensure immutability of the data structures.
    """
    try:
        len(list_obj)
        return list_obj
    except TypeError:
        return (list_obj,)


def larger_than_zero_error(value: int, arg_name: str):
    """
    Check if the value is larger than zero.
    """
    if value <= 0:
        raise ValueError(f"{arg_name} has to be larger than 0.")
    return value


def no_name_error(name: str) -> str:
    """
    Check if the name is not empty.
    """
    if len(name) == 0:
        raise ValueError("Names of Circuit objects need to have at least one letter.")
    return name.lower()


def nr_of_qubits_error(qubits: tuple[tuple[int, ...], ...], values: ValidationInfo):
    """
    Check if the number of qubits is consistent with the length of the pauli string.

    This `field_validator` is used for objects that include a `pauli` field.
    """

    pauli = retrieve_field("pauli", values)

    if len(qubits) != len(pauli):
        raise ValueError(
            f"Number of qubits: {len(qubits)} does not match the length of the"
            f" pauli string: {len(pauli)}."
        )

    return qubits


def pauli_error(pauli_str: str) -> str:
    """
    Check if the pauli is a valid pauli string.
    """
    valid_pauli_chars = ["X", "Y", "Z"]

    def error_message(input_pauli: str) -> ValueError:
        return ValueError(
            f"Invalid pauli: {input_pauli}. Must be one of {valid_pauli_chars}."
        )

    for p in pauli_str:
        if p not in valid_pauli_chars:
            raise error_message(p)
    return pauli_str


def retrieve_field(name: str, info: ValidationInfo):
    """
    Retrieve a field from the info object.
    """
    message = (
        "This is most probably due to a previous validation error. Check, that the "
        "other fields in the data class are set correctly."
    )

    try:
        result = info.data[name]
        if result is None:
            raise ValueError(f"Field {name} is None. " + message)
        return result
    except Exception as exc:
        raise ValueError(
            f"Field {name} not found in {info.config['title']} object. " + message
        ) from exc


def uuid_error(uuid_str: str) -> str:
    """
    Check if the uuid is a valid uuid.
    """
    try:
        uuid_obj = uuid.UUID(uuid_str)
    except ValueError as exc:
        raise ValueError(f"Invalid uuid: {uuid_str}. UUID must be version 4.") from exc

    if uuid_obj.version != 4:
        raise ValueError(f"Invalid uuid: {uuid_str}. UUID must be version 4.")

    return uuid_str



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from abc import ABC
from typing import Any
from pydantic.dataclasses import dataclass


@dataclass(frozen=True)
class AbstractValidityCheck(ABC):
    """Abstract dataclass to store the results of a check.

    Parameters
    ----------
    output: Any
        Any output that is relevant to the check. It has to have a __len__ method to be
        used to determine if the check is valid.

    Properties
    ----------
    valid: bool
        True if the check is valid (i.e., no issues found), False otherwise.
    message: str
        A message indicating the result of the check. It will be empty if the check is
        valid, otherwise it will contain a message describing the issue. This must be
        implemented by subclasses.
    """

    output: Any

    def __new__(cls, *args, **kwargs):  # pylint: disable=unused-argument
        if cls == AbstractValidityCheck:
            raise TypeError("Cannot instantiate abstract ValidityCheck class.")
        return super().__new__(cls)

    def __str__(self):
        out = f"{self.__class__.__name__}: valid = {self.valid}\n"
        if not self.valid:
            out += f"message: {self.message}\n"
            out += "output: \n"
            out += str(self.output) + "\n"
        return out.rstrip("\n")  # Remove the last newline character for cleaner output

    @property
    def valid(self) -> bool:
        """
        Determine if the check passed based on the output.
        """
        return len(self.output) == 0

    @property
    def message(self) -> str:
        """
        A message indicating the result of the check. It will be empty if the check is
        valid, otherwise it will contain a message describing the issue.
        """
        raise NotImplementedError(
            "Subclasses must implement the 'message' property to provide a message."
        )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Circuit, Stabilizer, PauliOperator
from loom.eka.utilities import Direction, DiagonalDirection
from loom.interpreter import InterpretationStep

from ..move_block import DetailedSchedule, direction_to_coord
from ..utilities import FourBodySchedule, find_relative_diagonal_direction
from ...code_factory import RotatedSurfaceCode

# pylint: disable=duplicate-code


def qubit_mapper(
    q: tuple[int, ...], move_direction: DiagonalDirection
) -> tuple[int, ...]:
    """
    Map qubit to appropriate diagonal qubit after movement.

    Parameters
    ----------
    q : tuple[int, ...]
        The original qubit coordinates.
    move_direction : DiagonalDirection
        The direction of movement.

    Returns
    -------
    tuple[int, ...]
        The new qubit coordinates after movement.
    """
    return tuple(
        map(
            sum,
            zip(
                q,
                direction_to_coord(move_direction, q[2]),
                strict=True,
            ),
        )
    )


def move_logicals_and_append_evolution_and_updates(
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    move_direction: DiagonalDirection,
) -> tuple[PauliOperator, PauliOperator]:
    """
    Move logical operators diagonally and append their evolution and updates
    to the interpretation step.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step.
    block : RotatedSurfaceCode
        The rotated surface code block whose logicals are to be moved.
    move_direction : DiagonalDirection
        The direction of movement.

    Returns
    -------
    tuple[PauliOperator, PauliOperator]
        The new logical X and Z operators.
    """
    # Find new logicals
    old_logical_x = block.logical_x_operators[0]
    new_logical_x = PauliOperator(
        pauli=old_logical_x.pauli,
        data_qubits=[
            qubit_mapper(dq, move_direction) for dq in old_logical_x.data_qubits
        ],
    )
    old_logical_z = block.logical_z_operators[0]
    new_logical_z = PauliOperator(
        pauli=old_logical_z.pauli,
        data_qubits=[
            qubit_mapper(dq, move_direction) for dq in old_logical_z.data_qubits
        ],
    )

    # Update evolution
    interpretation_step.logical_x_evolution[new_logical_x.uuid] = (old_logical_x.uuid,)
    interpretation_step.logical_z_evolution[new_logical_z.uuid] = (old_logical_z.uuid,)

    # Update logical updates
    interpretation_step.update_logical_operator_updates_MUT(
        "X",
        new_logical_x.uuid,
        (),
        inherit_updates=True,
    )
    interpretation_step.update_logical_operator_updates_MUT(
        "Z", new_logical_z.uuid, (), inherit_updates=True
    )

    return new_logical_x, new_logical_z


def generate_teleportation_finalization_circuit_with_updates(
    interpretation_step: InterpretationStep,
    new_block: RotatedSurfaceCode,
    anc_qubits_to_init: dict[str, list[tuple[int, ...]]],
    teleportation_qubit_pairs: list[tuple[tuple[int, ...], tuple[int, ...]]],
) -> Circuit:
    """
    Generate the circuit that finalizes the teleportation operation. This includes
    the measurement of the data qubits and the updating of the necessary stabilizers
    and logical operators based on the measurement outcomes to complete the
    teleportation without needing to apply any gates.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step.
    new_block : RotatedSurfaceCode
        The new block after the operation.
    anc_qubits_to_init : dict[str, list[tuple[int, ...]]]
        A dictionary containing the ancilla qubits to initialize in the X and Z basis.
    teleportation_qubit_pairs : list[tuple[tuple[int, ...], tuple[int, ...]]]
        The list of teleportation qubit pairs. The first qubit is the ancilla qubit
        and the second is the data qubit.

    Returns
    -------
    Circuit
        The circuit that finalizes the teleportation operation
    """

    teleportation_circ_seq = [[]]
    for corrected_qubit, measured_qubit in teleportation_qubit_pairs:
        # Obtain the necessary channels and cbit
        cbit = interpretation_step.get_new_cbit_MUT(f"c_{measured_qubit}")
        cbit_channel = interpretation_step.get_channel_MUT(
            f"{cbit[0]}_{cbit[1]}", channel_type="classical"
        )
        measured_qubit_channel = interpretation_step.get_channel_MUT(measured_qubit)

        # Determine the operations based on the initialization basis of the
        # ancilla qubit
        if corrected_qubit in anc_qubits_to_init["X"]:
            measure_op_name = "measure_z"
            stabs_to_update_pauli = "Z"
        elif corrected_qubit in anc_qubits_to_init["Z"]:
            measure_op_name = "measure_x"
            stabs_to_update_pauli = "X"
        else:
            raise ValueError("The ancilla qubit was not found in the initialization.")

        # Define the circuits that measure the data qubit and operate on the ancilla
        # qubit based on the measurement
        meas_circ = Circuit(
            measure_op_name,
            channels=[measured_qubit_channel, cbit_channel],
        )

        # Append the circuits to the list
        teleportation_circ_seq[0].append(meas_circ)

        # Find which qubit of the two is in the block and will be associated with
        # some stabilizers

        # Do the same for the logical operator (it's going to be only one operator)
        logical_ops = (
            new_block.logical_z_operators
            if stabs_to_update_pauli == "Z"
            else new_block.logical_x_operators
        )
        logical_ops_involved = [
            op for op in logical_ops if corrected_qubit in op.data_qubits
        ]
        for op in logical_ops_involved:
            interpretation_step.update_logical_operator_updates_MUT(
                stabs_to_update_pauli, op.uuid, (cbit,), False
            )

        # Find all appropriate stabilizers
        stabs_to_update = [
            stab
            for stab in new_block.stabilizers
            if stab.pauli[0] == stabs_to_update_pauli  # correct flavor
            and corrected_qubit in stab.data_qubits  # qubit part of stabilizer
        ]

        # Append the Cbit on the updates of the stabilizers
        for stab in stabs_to_update:
            current_upd = interpretation_step.stabilizer_updates.get(stab.uuid, ())
            interpretation_step.stabilizer_updates[stab.uuid] = current_upd + (cbit,)

    # Compile the teleportation circuit finalization
    teleportation_circuit_finalization = Circuit(
        "teleportation finalization",
        circuit=teleportation_circ_seq,
    )

    return teleportation_circuit_finalization


def get_final_block_syndrome_measurement_cnots_circuit(
    block: RotatedSurfaceCode,
    interpretation_step: InterpretationStep,
    is_swap_then_qec: bool = False,
    idle_side_directions: DiagonalDirection | None = None,
    anc_qubits_to_init: dict[str, list[tuple[int, ...]]] | None = None,
) -> Circuit:
    # pylint: disable=too-many-locals, too-many-branches
    """
    Generates fault-tolerant CNOT scheduling for syndrome measurement of
    the final block. The block is divided into regions as explained below.

    Examples: a 7x13 block with three lines corresponding to three ways
    for dividing the block::

                3 * * * * * 4
                *         . *
                * .     .   *
                *   . .     *
                *   . .     *
                * .     .   *
                2         . *
                * .         o
                *   .       *
                *     .     *
                *       .   *
                *         . *
                1 * * * * * *

    - Region 1: draw a diagonal line from qubit o towards corner 3, note that the \
    line doesn't pass through corner 3. Qubit o is the qubit shifted by 1 from the \
    opposite of corner 2 away from corner 4. Region 2 is defined as the region \
    containing corner 4.
    - Region 2: draw a diagonal line from corner 2 (middle-edge corner) to corner 4 \
    (short end corner). Region 2 is defined as the region containing corner 3 \
    (angle corner).
    - Region 3: draw a diagonal line from corner 2 to the non-topological geometric \
    corner. Region 3 is defined as the region containing corner 1.
    - Region 4: qubits not contained in region 1, 2, or 3.

    The scheduling algorithm is the following.

    - Z stabilizers outside region 3 are scheduled so that Z errors propagate \
    perpendicular to the edge connecting corners 3 and 4.
    - Z stabilizers inside region 3 are scheduled opposite to Z stabilizers \
    outside the region.
    - X stabilizers inside regions 1 and 2 are scheduled opposite to the schedule of Z \
    stabilizers outside region 3.
    - X stabilizers outside regions 1 and 2 are scheduled opposite to the schedule of \
    X stabilizers inside regions 1 and 2.

    Parameters
    ----------
    block : RotatedSurfaceCode
        The final block after the y_wall_out operation.
    interpretation_step : InterpretationStep
        The interpretation step for the circuit generation.
    is_swap_then_qec: bool
        True if the syndrome measurement round is a SWAP-then-QEC round
    idle_side_directions: DiagonalDirection
        The directions that the idling side half of the block will move.
    anc_qubits_to_init: dict[str, list[tuple[int, ...]]]
        The ancilla qubits to initialize in the second SWAP-then-QEC round.

    Returns
    -------
    Circuit
        The generated CNOT circuit for the syndrome measurement of the final block.

    """
    config, pivot_corners = block.config_and_pivot_corners
    if config != 3:
        raise ValueError(
            "The final block of Y-wall-out must have type-3 corner configuration."
        )

    short_end_corner = pivot_corners[3]
    top_left_corner = block.upper_left_qubit
    is_horizontal = block.is_horizontal

    # The starting qubit direction is the direction of the short-end corner
    starting_diag_direction = find_relative_diagonal_direction(
        top_left_corner, short_end_corner
    )

    # Schedules
    # Z schedule is determined based on the orientation of the short edge
    region_124_z_schedule = FourBodySchedule.Z if is_horizontal else FourBodySchedule.N
    region_3_z_schedule = region_124_z_schedule.opposite_schedule()
    region_12_x_schedule = region_124_z_schedule.opposite_schedule()
    region_34_x_schedule = region_12_x_schedule.opposite_schedule()

    # Determine cnot sequence, require 5 time steps
    cnots = [[] for _ in range(5)]

    if is_swap_then_qec:
        vec_from_final_pos = direction_to_coord(idle_side_directions, 0)
        for q in block.data_qubits:
            # q (final position) is in sublattice 0, while its current position is in
            # sublattice 1
            q_current_pos = tuple(map(sum, zip(q, vec_from_final_pos, strict=True)))

            if q in anc_qubits_to_init["X"]:
                cnot_pair = (q, q_current_pos)
            elif q in anc_qubits_to_init["Z"]:
                cnot_pair = (q_current_pos, q)
            else:
                raise ValueError("The ancilla qubit was not found.")
            cnots[0].append(cnot_pair)

    for stab in block.stabilizers:
        pauli_type = stab.pauli_type
        (
            is_in_region_12,
            is_in_region_3,
            is_on_boundary_12,
            _,
            _,
            boundary_direction,
        ) = find_stabilizer_position_in_final_block(block, stab, pivot_corners)
        if pauli_type == "Z":
            schedule = region_3_z_schedule if is_in_region_3 else region_124_z_schedule
        else:
            schedule = region_12_x_schedule if is_in_region_12 else region_34_x_schedule

        detailed_schedule = DetailedSchedule.from_schedule_and_direction(
            schedule, starting_diag_direction
        )

        if len(stab.data_qubits) == 4:
            data_qubits = detailed_schedule.get_stabilizer_qubits(stab)
        else:
            data_qubits = detailed_schedule.get_stabilizer_qubits(
                stab, boundary_direction
            )

        for i, data_qubit in enumerate(data_qubits):
            if (is_swap_then_qec and i == 0) or data_qubit is None:
                continue
            if pauli_type == "Z":
                # If the stabilizer is Z, cnot from data qubit to ancilla
                cnot_pair = (data_qubit, stab.ancilla_qubits[0])
            elif pauli_type == "X":
                # If the stabilizer is X, cnot from ancilla to data qubit
                cnot_pair = (stab.ancilla_qubits[0], data_qubit)
            else:
                raise ValueError("Unknown stabilizer type.")

            # Determine the cnot layer
            layer = (
                i + 1
                if (
                    i > 0
                    and (
                        (
                            not is_in_region_12 and not is_on_boundary_12
                        )  # bulk of region 3 and 4
                        or (
                            not is_in_region_12 and is_on_boundary_12 and i == 3
                        )  # boundary of region 4
                    )
                )
                else i
            )
            cnots[layer].append(cnot_pair)

    # Generate the circuit containing the cnots
    cnots_circuit = Circuit(
        name="Final block syndrome measurement CNOT circuit",
        circuit=[
            [
                Circuit(
                    "cx",
                    channels=[
                        interpretation_step.get_channel_MUT(q) for q in qubit_pair
                    ],
                )
                for qubit_pair in cnot_slice
            ]
            for cnot_slice in cnots
        ],
    )

    return cnots_circuit


def find_stabilizer_position_in_final_block(
    block: RotatedSurfaceCode,
    stab: Stabilizer,
    pivot_corners: tuple[
        tuple[int, ...], tuple[int, ...], tuple[int, ...], tuple[int, ...]
    ],
) -> tuple[bool, bool, bool, bool, bool, Direction | None]:
    # pylint: disable=too-many-locals
    """
    Find stabilizer position with respect to the regions of the final block

    Parameters
    ----------
    block: RotatedSurfaceCode
        The final block after the y_wall_out operation.
    stab: Stabilizer
        The input stabilizer.
    pivot_corners:
        tuple[tuple[int, ...], tuple[int, ...], tuple[int, ...], tuple[int, ...]]
        Four topological corners.

    Returns
    -------
    tuple[bool, bool, bool, Direction | None]
        True if the stabilizer is inside region 1 or 2
        True if the stabilizer is inside region 3
        True if the stabilizer is inside region 4 and next to region 1 or 2
        True if the stabilizer in on the wall
        True if the stabilizer in the Hadamard side
        Direction of the boundary if the input stabilizer is a weight-2 stabilizer

    """

    long_end_corner, middle_corner, _, short_end_corner = pivot_corners
    is_horizontal = block.is_horizontal
    long_edge_idx = 0 if is_horizontal else 1  # coordinate index along the long side
    short_edge_idx = 1 if is_horizontal else 0  # coordinate index along the short side

    def is_in_region_1(q) -> bool:
        """Determine if the input qubit is inside region 1
        (including the boundaries of the region). This is done by checking
        whether the distance along the long side of the block between the input
        qubit and the qubit o is at least the distance along the
        short side between the same two qubits.
        """
        # Compare middle_corner and short_end_corner along the short edge to determine
        # if the qubit is on the left or right of middle_corner.
        # Distance is calculated with respect to short_end_corner because it has the
        # same (short-edge) coordinate as qubit o.
        dshort = (
            q[short_edge_idx] - short_end_corner[short_edge_idx]
            if middle_corner[short_edge_idx] > short_end_corner[short_edge_idx]
            else short_end_corner[short_edge_idx] - q[short_edge_idx]
        )
        # Compare middle_corner and long_end_corner along the long edge to determine
        # if the qubit is on the top or bottom of middle_corner.
        # Distance is calculated with respect to middle_corner because the (long-edge)
        # coordinate of qubit o can be deduced from it.
        dlong = (
            middle_corner[long_edge_idx] - q[long_edge_idx] + 1
            if middle_corner[long_edge_idx] < long_end_corner[long_edge_idx]
            else q[long_edge_idx] - middle_corner[long_edge_idx] + 1
        )
        return 0 <= dshort <= dlong and dlong >= 0

    def is_in_region_2(q) -> bool:
        """Determine if the input qubit is inside region 2
        (including the boundaries of the region). This is done by checking
        whether the distance along the long side of the block between the input
        qubit and the middle_corner is at least the distance along the
        short side between the same two qubits.
        """
        # Compare middle_corner and short_end_corner along the short edge to determine
        # if the qubit is on the left or right of middle_corner.
        # Distance is calculated with respect to middle_corner
        dshort = (
            q[short_edge_idx] - middle_corner[short_edge_idx]
            if middle_corner[short_edge_idx] < short_end_corner[short_edge_idx]
            else middle_corner[short_edge_idx] - q[short_edge_idx]
        )
        # Compare middle_corner and long_end_corner along the long edge to determine
        # if the qubit is on the top or bottom of middle_corner.
        # Distance is calculated with respect to middle_corner
        dlong = (
            middle_corner[long_edge_idx] - q[long_edge_idx]
            if middle_corner[long_edge_idx] < long_end_corner[long_edge_idx]
            else q[long_edge_idx] - middle_corner[long_edge_idx]
        )
        return 0 <= dshort <= dlong and dlong >= 0

    def is_in_region_3(q) -> bool:
        """Determine if the input qubit is inside region 3
        (including the boundaries of the region). This is done by checking
        whether the distance along the long side of the block between the input
        qubit and the qubit o is at least the distance along the
        short side between the same two qubits.
        """
        # Compare middle_corner and short_end_corner along the short edge to determine
        # if the qubit is on the left or right of middle_corner.
        # Distance is calculated with respect to middle_corner
        dshort = (
            q[short_edge_idx] - middle_corner[short_edge_idx]
            if middle_corner[short_edge_idx] < short_end_corner[short_edge_idx]
            else middle_corner[short_edge_idx] - q[short_edge_idx]
        )
        # Compare middle_corner and long_end_corner along the long edge to determine
        # if the qubit is on the top or bottom of middle_corner.
        # Distance is calculated with respect to middle_corner
        # Note that region 3 is on the opposite side (along the long edge) of region 2
        # with respect to middle_corner.
        dlong = (
            q[long_edge_idx] - middle_corner[long_edge_idx]
            if middle_corner[long_edge_idx] < long_end_corner[long_edge_idx]
            else middle_corner[long_edge_idx] - q[long_edge_idx]
        )
        return 0 <= dshort <= dlong and dlong >= 0

    dq_in_region_1 = [dqubit for dqubit in stab.data_qubits if is_in_region_1(dqubit)]
    dq_in_region_2 = [dqubit for dqubit in stab.data_qubits if is_in_region_2(dqubit)]
    dq_in_region_3 = [dqubit for dqubit in stab.data_qubits if is_in_region_3(dqubit)]

    is_stab_in_region_1 = len(dq_in_region_1) == len(stab.data_qubits)
    is_stab_in_region_2 = len(dq_in_region_2) == len(stab.data_qubits)
    is_stab_in_region_3 = len(dq_in_region_3) == len(stab.data_qubits)
    is_stab_in_region_4 = not (
        is_stab_in_region_1 or is_stab_in_region_2 or is_stab_in_region_3
    )

    is_stab_in_region_12 = is_stab_in_region_1 or is_stab_in_region_2

    is_stab_on_boundary_region_1 = (len(stab.data_qubits) - len(dq_in_region_1)) == 1
    is_stab_on_boundary_region_2 = (len(stab.data_qubits) - len(dq_in_region_2)) == 1
    is_stab_in_region_4_and_on_boundary_region_12 = is_stab_in_region_4 and (
        is_stab_on_boundary_region_1 or is_stab_on_boundary_region_2
    )

    # Find the stabilizer's position relative to the wall
    top_left_qubit = max(stab.data_qubits, key=lambda x: -x[0] - x[1])
    bottom_right_qubit = max(stab.data_qubits, key=lambda x: x[0] + x[1])
    compare_qubit = (
        top_left_qubit
        if middle_corner[long_edge_idx] < long_end_corner[long_edge_idx]
        else bottom_right_qubit
    )
    is_on_wall = middle_corner[long_edge_idx] == compare_qubit[long_edge_idx]
    is_on_hadamard_side = (
        compare_qubit[long_edge_idx] > middle_corner[long_edge_idx]
        if long_end_corner[long_edge_idx] > middle_corner[long_edge_idx]
        else compare_qubit[long_edge_idx] < middle_corner[long_edge_idx]
    )

    # Find the boundary direction of weight-2 stabilizer
    boundary = None
    weight = len(stab.pauli)
    if weight == 2:
        qubit_directions = [
            set(
                direction
                for direction in Direction
                if dqubit in block.boundary_qubits(direction)
            )
            for dqubit in stab.data_qubits
        ]
        boundary = list(qubit_directions[0].intersection(qubit_directions[1]))[0]

    return (
        is_stab_in_region_12,
        is_stab_in_region_3,
        is_stab_in_region_4_and_on_boundary_region_12,
        is_on_wall,
        is_on_hadamard_side,
        boundary,
    )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# Populate `__revision__` variable holding the actual git revision. The procedure is as
# follows: (1) Suppose we are running from an editable install. Then the actual revision
# will be known by `git rev-parse HEAD`. (2) If this fails, then we probably run the
# packaged version of Loom. In this case, the `./scripts/update-revision.py` did its job
# and the `_revision.py` module is available.  So we re-export `__revision__` from
# there. (3) If failed, then something unexpected has happened, we set `__revision__` to
# None.

try:
    import subprocess
    from pathlib import Path

    __revision__ = (
        subprocess.check_output(
            ["git", "rev-parse", "HEAD"],
            cwd=(Path(__file__).parent.parent.parent / ".git").resolve(),
            stderr=subprocess.DEVNULL,
        )
        .decode()
        .strip()
    )
except Exception:  # pylint: disable=broad-exception-caught
    try:
        from ._revision import __revision__
    except ImportError:
        __revision__ = None



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Circuit, PauliOperator, Stabilizer
from loom.eka.operations import Merge
from loom.eka.utilities import Direction, Orientation
from loom.interpreter import InterpretationStep

from ..code_factory import RepetitionCode


# pylint: disable=too-many-locals
def merge_consistency_check(
    interpretation_step: InterpretationStep, operation: Merge
) -> tuple[RepetitionCode, RepetitionCode]:
    """
    Check the consistency of the merge operation.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        InterpretationStep containing the blocks to merge.
    operation : Merge
        Merge operation description.

    Returns
    -------
    tuple[RepetitionCode,RepetitionCode]
        Blocks to merge.
    """

    name_1, name_2 = operation.input_blocks_name
    block_1 = interpretation_step.get_block(name_1)
    block_2 = interpretation_step.get_block(name_2)

    # Check valid Block type
    if not isinstance(block_1, RepetitionCode) or not isinstance(
        block_2, RepetitionCode
    ):
        wrong_types = tuple(
            set((type(block_1), type(block_2))).difference({RepetitionCode})
        )
        raise TypeError(
            f"This merge operation is not supported for {tuple(wrong_types)} blocks."
        )

    # Ensure block_1 is left and block_2 is right
    if (
        block_1.boundary_qubits(Direction.LEFT)[0]
        > block_2.boundary_qubits(Direction.LEFT)[0]
    ):
        block_1, block_2 = block_2, block_1

    # Check that Blocks have same type of stabilizers
    if block_1.check_type != block_2.check_type:
        raise TypeError(
            "Cannot merge blocks with different check types: "
            f"{block_1.check_type} and {block_2.check_type}"
        )

    # Check that Blocks do not overlap
    left_boundary_block_1 = block_1.boundary_qubits(Direction.LEFT)[0]
    distance_block_1 = len(block_1.data_qubits)
    left_boundary_block_2 = block_2.boundary_qubits(Direction.LEFT)[0]

    if left_boundary_block_2 < left_boundary_block_1 + distance_block_1:
        raise ValueError("Cannot merge blocks that overlap.")

    # Check that orientation is horizontal
    if operation.orientation != Orientation.HORIZONTAL:
        raise ValueError(
            "Repetition code does not support merging in the vertical orientation."
        )

    return block_1, block_2


def get_new_data_qubits_info(
    blocks: list[RepetitionCode],
) -> list[tuple[int, int]]:
    """Get the new data qubits to be added in the merged block.

    Parameters
    ----------
    blocks : list[RepetitionCode]
        List of blocks (capped at two) to be merged.

    Returns
    -------
    list[tuple[int, int]]
        List of new data qubits to be added.
    """

    # Extract blocks
    block_1, block_2 = blocks

    # Extract boundaries
    right_boundary_block_1 = block_1.boundary_qubits(Direction.RIGHT)[0]
    left_boundary_block_2 = block_2.boundary_qubits(Direction.LEFT)[0]

    # Define new data qubits
    new_data_qubits = [
        (i, 0) for i in range(right_boundary_block_1 + 1, left_boundary_block_2)
    ]

    return new_data_qubits


def find_merge_circuit(
    interpretation_step: InterpretationStep,
    check_type: str,
    new_data_qubits: list[tuple[int, int]],
    circuit_name: str,
) -> Circuit:
    """
    Generates the circuit required to reset the new data that get added during the
    merge operation.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the blocks to be merged.
    check_type : str
        Type of stabilizers in the codes.
    new_data_qubits : list[tuple[int,int]]
        List of new data qubits to reset.
    circuit_name : str
        Name of the circuit.

    Returns
    -------
    Circuit
        Circuit to reset the new data and ancilla qubits.
    """

    # Select reset state based on stabilizer type
    if check_type == "X":
        reset_state = "0"
    else:
        reset_state = "+"

    # Reset sequence
    reset_sequence = [
        [
            Circuit(
                name=f"reset_{reset_state}",
                channels=interpretation_step.get_channel_MUT(q, "quantum"),
            )
            for q in new_data_qubits
        ]
    ]

    # Circuit for the merge operation
    merge_circuit = Circuit(
        name=circuit_name,
        circuit=reset_sequence,
    )
    return merge_circuit


def find_new_stabilizers(
    blocks: list[RepetitionCode], check_type: str
) -> list[Stabilizer]:
    """Find the new stabilizers for the merged block.

    Parameters
    ----------
    blocks : list[RepetitionCode]
        List of blocks (capped at two) to be merged.
    check_type : str
        Type of stabilizers in the codes.

    Returns
    -------
    list[Stabilizer]
        New stabilizers for the merged block.
    """

    # Extract blocks
    block_1, block_2 = blocks

    # Extract boundaries
    right_boundary_block_1 = block_1.boundary_qubits(Direction.RIGHT)[0]
    left_boundary_block_2 = block_2.boundary_qubits(Direction.LEFT)[0]

    # Generate added stabilizers
    add_stabilizers = [
        Stabilizer(
            pauli=check_type * 2,
            data_qubits=[(i, 0), (i + 1, 0)],
            ancilla_qubits=[(i, 1)],
        )
        for i in range(right_boundary_block_1, left_boundary_block_2)
    ]

    # Generate new stabilizers
    new_stabilizers = (
        list(block_1.stabilizers) + add_stabilizers + list(block_2.stabilizers)
    )

    return new_stabilizers


def get_logical_operator_and_evolution(  # pylint: disable=too-many-locals
    blocks: list[RepetitionCode],
    check_type: str,
    new_data_qubits: list[tuple[int, int]],
) -> tuple[list[list[PauliOperator]], list[dict]]:
    """
    Get the new logical operators and evolution for the merged block.

    Parameters
    ----------
    blocks : list[RepetitionCode]
        List of blocks (capped at two) to be merged.
    check_type : str
        Type of stabilizers in the codes.
    new_data_qubits : list[tuple[int,int]]
        List of new data qubits to be added.

    Returns
    -------
    tuple[list[list[PauliOperator]], list[dict]]
        New logical operators and evolution for the merged block.
    """

    block_1, block_2 = blocks

    other_check_type = "X" if check_type == "Z" else "Z"

    # Extract old long and short operators
    old_logs_1 = [block_1.logical_x_operators[0], block_1.logical_z_operators[0]]
    old_short_log_1, old_long_log_1 = sorted(
        old_logs_1, key=lambda x: len(x.data_qubits)
    )
    old_logs_2 = [block_2.logical_x_operators[0], block_2.logical_z_operators[0]]
    _, old_long_log_2 = sorted(old_logs_2, key=lambda x: len(x.data_qubits))

    new_long_data_qubits = (
        list(block_1.data_qubits) + new_data_qubits + list(block_2.data_qubits)
    )
    new_long_log = PauliOperator(
        pauli=other_check_type * len(new_long_data_qubits),
        data_qubits=new_long_data_qubits,
    )

    # We impose Block 1 to be the left one in the main merge function
    left_boundary_block_1 = block_1.boundary_qubits(Direction.LEFT)
    new_short_log, stabs_required = block_1.get_shifted_equivalent_logical_operator(
        left_boundary_block_1
    )

    # New logicals
    new_logs = [[new_long_log], [new_short_log]]

    long_log_evolution = {new_long_log.uuid: (old_long_log_1.uuid, old_long_log_2.uuid)}
    if stabs_required:
        id_stabs_required = [stab.uuid for stab in stabs_required]
        short_log_evolution = {
            new_short_log.uuid: tuple([old_short_log_1.uuid] + id_stabs_required)
        }
    else:
        short_log_evolution = {}

    # Evolution
    log_evolution = [long_log_evolution, short_log_evolution]

    return new_logs, log_evolution


def merge(  # pylint: disable=too-many-locals
    interpretation_step: InterpretationStep,
    operation: Merge,
    same_timeslice: bool,
    debug_mode: bool,
):
    """
    Merge two Repetition Code chains.

    The algorithm is the following:

    - A.) DATA QUBITS

        - A.1) Find the new data qubits to be added in the merged block

    - B.) CIRCUIT

        - B.1) Find the circuit for the merge operation
        - B.2) Add the circuit to the interpretation step

    - C.) - STABILIZERS
        - C.1) Find the new stabilizers for the merged block

    - D.) LOGICAL OPERATORS

        - D.1) Find new logical operator and evolution
        - D.2) Update the logical operator history

    - E.) NEW BLOCK AND NEW INTERPRETATION STEP

        - E.1) Create the new block
        - E.2) Update the block history

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the blocks to merge.
    operation : Merge
        Merge operation description.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block.

    Returns
    -------
    InterpretationStep
        Interpretation step after the merge operation.
    """

    # Merge consistency check
    block_1, block_2 = merge_consistency_check(interpretation_step, operation)

    # Extract check type
    check_type = block_1.check_type

    # A) DATA QUBITS
    # A.1) Find the new data qubits to be added in the merged block
    new_data_qubits = get_new_data_qubits_info([block_1, block_2])

    # B) CIRCUIT
    # B.1) Find the circuit for the merge operation
    circuit_name = (
        f"merge {block_1.unique_label} and "
        f"{block_2.unique_label} into {operation.output_block_name}"
    )
    merge_circuit = find_merge_circuit(
        interpretation_step, check_type, new_data_qubits, circuit_name
    )

    # B.2) Add the circuit to the interpretation step
    interpretation_step.append_circuit_MUT(merge_circuit, same_timeslice)

    # C) STABILIZERS
    # C.1) Find the new stabilizers for the merged block
    new_stabilizers = find_new_stabilizers([block_1, block_2], check_type)

    # D) LOGICAL OPERATORS
    # D.1) Find new logical operator and evolution
    new_logs, log_evolution = get_logical_operator_and_evolution(
        [block_1, block_2], check_type, new_data_qubits
    )

    # D.2) Update the logical operator history
    ordering = 1 if check_type == "Z" else -1
    new_log_x_ops, new_log_z_ops = new_logs[::ordering]
    logical_x_evolution, logical_z_evolution = log_evolution[::ordering]
    interpretation_step.logical_x_evolution.update(logical_x_evolution)
    interpretation_step.logical_z_evolution.update(logical_z_evolution)
    # Inherit updates for the new operators
    for op in new_log_x_ops:
        interpretation_step.update_logical_operator_updates_MUT("X", op.uuid, (), True)
    for op in new_log_z_ops:
        interpretation_step.update_logical_operator_updates_MUT("Z", op.uuid, (), True)

    # E) NEW BLOCK
    # E.1) Create the new block
    new_block = RepetitionCode(
        stabilizers=new_stabilizers,
        logical_x_operators=new_log_x_ops,
        logical_z_operators=new_log_z_ops,
        unique_label=operation.output_block_name,
        skip_validation=not debug_mode,
    )

    # E.2) Update the block history
    interpretation_step.update_block_history_and_evolution_MUT(
        new_blocks=(new_block,),
        old_blocks=(block_1, block_2),
    )

    return interpretation_step



import subprocess
from pathlib import Path

try:
    revision = (
        '"'
        + (
            subprocess.check_output(
                ["git", "rev-parse", "HEAD"],
                cwd=Path(".git").resolve(),
            )
            .decode()
            .strip()
        )
        + '"'
    )

except Exception:
    revision = "None # Build system failed to determine the revision"

with open(Path("src") / "loom" / "_revision.py", "w") as f:
    f.write(f'"""Records Loom source revision as seen by the distribution system.\n')
    f.write(f'(AUTOGENERATED by ./scripts/update-revision.py) """\n\n')
    f.write(f"__revision__ = {revision}\n")



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=duplicate-code
from loom.eka import Circuit, PauliOperator
from loom.eka.operations import Grow
from loom.eka.utilities import Direction, Orientation
from loom.interpreter import InterpretationStep
from loom.interpreter.applicator.generate_syndromes import generate_syndromes

from ..code_factory import RotatedSurfaceCode


def grow(  # pylint: disable=too-many-branches, too-many-statements, too-many-locals
    interpretation_step: InterpretationStep,
    operation: Grow,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Grow a Block in the specified direction.

    The algorithm is the following:

    - A.) Create the new piece of surface code (the "extra" part of the grown block)

        - A.1) Find the corner of the additional piece
        - A.2) Find the Pauli flavour of the upper left 4-body for the new piece
        - A.3) Create the new 4-body stabilizers
        - A.4) Create the new 2-body stabilizers
        - A.5) Identify the stabilizers to remove and the ones to replace them with
        - A.6) Add the new stabilizers to the stabilizer_evolution

    - B.) Create the reset circuit

        - B.1) Reset all data qubits involved in the growth in the right basis
        - B.2) Create empty syndromes for the relevant stabilizers

    - C.) Associate the new stabilizers to their respective syndrome circuits

        - C.1) Recover the syndrome circuits by name
        - C.2) Update the stabilizer to circuit mapping

    - D.) Create the new logical operators and keep track of the evolution

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the block to grow.
    operation : Grow
        Grow operation description.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block

    Returns
    -------
    InterpretationStep
        Interpretation step after the grow operation.
    """

    # Get some useful information from the block
    block = interpretation_step.get_block(operation.input_block_name)
    if not isinstance(block, RotatedSurfaceCode):
        raise TypeError(
            f"The grow operation is not supported for {type(block)} blocks."
        )
    stabilizers = block.stabilizers
    direction = operation.direction
    length = operation.length
    old_x_logicals = block.logical_x_operators
    old_z_logicals = block.logical_z_operators

    is_horizontal = direction in (Direction.LEFT, Direction.RIGHT)
    dx = length + 1 if is_horizontal else block.size[0]
    dz = length + 1 if not is_horizontal else block.size[1]
    weight_4_x_schedule = block.weight_4_x_schedule
    weight_4_z_schedule = block.weight_4_z_schedule

    # A) - Create the new piece of surface code
    #   A.1) - Find the new corner of the block
    new_upper_left_position = list(block.upper_left_qubit[:2])
    match direction:
        case Direction.LEFT:
            new_upper_left_position[0] = new_upper_left_position[0] - length
        case Direction.RIGHT:
            new_upper_left_position[0] = new_upper_left_position[0] + (
                block.size[0] - 1
            )
        case Direction.TOP:
            new_upper_left_position[1] = new_upper_left_position[1] - length
        case Direction.BOTTOM:
            new_upper_left_position[1] = new_upper_left_position[1] + (
                block.size[1] - 1
            )

    #   A.2) - For the newly created stabilizers, find the Pauli flavour of the upper
    #   leftmost 4-body stabilizer.
    #   The pauli string of the upper left corner is conserved if the sum of the
    #   size of the block in the direction of growth and the length of growth is
    #   even. Otherwise, the Pauli string is flipped.
    if (
        length * (direction in (Direction.TOP, Direction.LEFT))
        + (block.size[0] - 1) * (direction == Direction.RIGHT)
        + (block.size[1] - 1) * (direction == Direction.BOTTOM)
    ) % 2 == 0:
        new_upper_left_4body_pauli = block.upper_left_4body_stabilizer.pauli
    else:
        new_upper_left_4body_pauli = (
            "ZZZZ" if block.upper_left_4body_stabilizer.pauli == "XXXX" else "XXXX"
        )

    #   A.3) - Create the new 4-body stabilizers
    new_upleft_4_body_stabs = RotatedSurfaceCode.generate_weight4_stabs(
        pauli=new_upper_left_4body_pauli,
        schedule=(
            weight_4_x_schedule
            if new_upper_left_4body_pauli == "XXXX"
            else weight_4_z_schedule
        ),
        start_in_top_left_corner=True,
        dx=dx,
        dz=dz,
        initial_position=new_upper_left_position,
    )
    new_rest_4_body_stabs = RotatedSurfaceCode.generate_weight4_stabs(
        pauli="XXXX" if new_upper_left_4body_pauli == "ZZZZ" else "ZZZZ",
        schedule=(
            weight_4_x_schedule
            if new_upper_left_4body_pauli != "XXXX"
            else weight_4_z_schedule
        ),
        start_in_top_left_corner=False,
        dx=dx,
        dz=dz,
        initial_position=new_upper_left_position,
    )
    new_4_body_stabs = new_upleft_4_body_stabs + new_rest_4_body_stabs

    #   A.4) - Create the new 2-body stabilizers
    new_top_left_is_xxxx = new_upper_left_4body_pauli == "XXXX"
    stab_left_right_is_x = block.x_boundary == Orientation.HORIZONTAL
    # Does the new left boundary start with a 2-body stabilizer on the first row?
    new_weight_2_stab_is_first_row = new_top_left_is_xxxx != stab_left_right_is_x
    stab_left_right = "XX" if stab_left_right_is_x else "ZZ"
    stab_top_bottom = "ZZ" if stab_left_right_is_x else "XX"
    # Left boundary
    if direction != Direction.RIGHT:
        if dz % 2 == 1:
            num_weight2_stabs = (dz - 1) / 2
        else:
            num_weight2_stabs = dz / 2 - (not new_weight_2_stab_is_first_row)
        stabs_left = RotatedSurfaceCode.generate_weight2_stabs(
            pauli=stab_left_right,
            initial_position=(
                new_upper_left_position[0],
                new_upper_left_position[1] + (not new_weight_2_stab_is_first_row),
            ),
            num_stabs=num_weight2_stabs,
            orientation=Orientation.VERTICAL,
            is_bottom_or_right=False,
        )
    else:
        stabs_left = []
    # Right boundary
    if direction != Direction.LEFT:
        # Does the new right boundary start with a 2-body stabilizer on the first row?
        new_right_first_row = new_weight_2_stab_is_first_row != (dx % 2)
        if dz % 2 == 1:
            num_weight2_stabs = (dz - 1) / 2
        else:
            num_weight2_stabs = dz / 2 - (not new_right_first_row)
        stabs_right = RotatedSurfaceCode.generate_weight2_stabs(
            pauli=stab_left_right,
            initial_position=(
                new_upper_left_position[0] + dx - 1,
                new_upper_left_position[1] + (not new_right_first_row),
            ),
            num_stabs=num_weight2_stabs,
            orientation=Orientation.VERTICAL,
            is_bottom_or_right=True,
        )
    else:
        stabs_right = []
    # Top boundary
    if direction != Direction.BOTTOM:
        # Does the new top boundary start with a 2-body stabilizer on the first column?
        num_weight2_stabs = dx // 2
        if dx % 2 == 1:
            num_weight2_stabs = (dx - 1) / 2
        else:
            num_weight2_stabs = dx / 2 - new_weight_2_stab_is_first_row
        stabs_top = RotatedSurfaceCode.generate_weight2_stabs(
            pauli=stab_top_bottom,
            initial_position=(
                new_upper_left_position[0] + new_weight_2_stab_is_first_row,
                new_upper_left_position[1],
            ),
            num_stabs=num_weight2_stabs,
            orientation=Orientation.HORIZONTAL,
            is_bottom_or_right=False,
        )
    else:
        stabs_top = []
    # Bottom boundary
    if direction != Direction.TOP:
        # Does the new bottom boundary start with a 2-body stabilizer on the first
        # column?
        bottom_first_col = new_weight_2_stab_is_first_row == (dz % 2)
        if dx % 2 == 1:
            num_weight2_stabs = (dx - 1) / 2
        else:
            num_weight2_stabs = dx / 2 - (not bottom_first_col)
        stabs_bottom = RotatedSurfaceCode.generate_weight2_stabs(
            pauli=stab_top_bottom,
            initial_position=(
                new_upper_left_position[0] + (not bottom_first_col),
                new_upper_left_position[1] + dz - 1,
            ),
            num_stabs=num_weight2_stabs,
            orientation=Orientation.HORIZONTAL,
            is_bottom_or_right=True,
        )
    else:
        stabs_bottom = []

    #   A.5) - Identify the stabilizers to remove and the ones to replace them with
    stabs_to_remove = [
        stab
        for stab in stabilizers
        if set(stab.data_qubits).issubset(set(block.boundary_qubits(direction)))
        and len(stab.data_qubits) == 2
    ]
    stabs_uuid_to_remove = [stab.uuid for stab in stabs_to_remove]
    # There is a 1-to-1 mapping between the stabilizers to remove and the new 4-body
    # stabilizers in stabs_uuid_replace. This mapping is used to update the
    # stabilizer_evolution dictionary
    stabs_uuid_replace = [
        stab.uuid
        for old_stab in stabs_to_remove
        for stab in new_4_body_stabs
        if set(old_stab.data_qubits).issubset(set(stab.data_qubits))
        and old_stab.pauli in stab.pauli
    ]
    final_to_initial_stab_map = {
        stabs_uuid_replace[i]: (uuid,) for i, uuid in enumerate(stabs_uuid_to_remove)
    }
    #   A.6) - Add the new stabilizers to the stabilizer_evolution
    interpretation_step.stabilizer_evolution.update(final_to_initial_stab_map)

    new_2_body_stabs = stabs_left + stabs_right + stabs_top + stabs_bottom
    new_stabilizers = (
        [stab for stab in stabilizers if stab not in stabs_to_remove]
        + new_2_body_stabs
        + new_4_body_stabs
    )

    # B) - Create the reset circuit
    #   B.1) - Reset all data qubits involved in the growth in the right basis
    if block.boundary_type(direction) == "X":
        # X boundary means Z stabilizers
        reset_state = "0"
    else:
        # Z boundary means X stabilizers
        reset_state = "+"
    new_data_qubits = tuple(
        set(
            q
            for stab in new_4_body_stabs
            for q in stab.data_qubits
            if q not in block.boundary_qubits(direction)
        )
    )
    existing_ancilla_qubits = tuple(
        q for stab in stabs_to_remove for q in stab.ancilla_qubits
    )
    reset_sequence = [
        [
            Circuit(
                name=f"reset_{reset_state}",
                channels=interpretation_step.get_channel_MUT(q, "quantum"),
            )
            for q in new_data_qubits
        ]
    ]
    grow_circuit = Circuit(
        name=f"grow {block.unique_label} by {length} to the {direction.value}",
        circuit=reset_sequence,
    )
    interpretation_step.append_circuit_MUT(grow_circuit, same_timeslice)

    #   B.2) - Create empty syndromes for the relevant stabilizers
    deterministic_stab_pauli = "X" if reset_state == "+" else "Z"
    # The new syndromes are created for the new 4-body stabilizers that are not morphed
    # from a 2-body. If they are, no need to account for corrections (it would be 0).
    relevant_stabs = [
        stab
        for stab in new_2_body_stabs + new_4_body_stabs
        if set(stab.pauli) == {deterministic_stab_pauli}
        and not all(a in existing_ancilla_qubits for a in stab.ancilla_qubits)
    ]
    new_syndromes = generate_syndromes(
        interpretation_step=interpretation_step,
        stabilizers=relevant_stabs,
        block=block,
        stab_measurements=[() for _ in relevant_stabs],
    )
    interpretation_step.append_syndromes_MUT(new_syndromes)

    # C) - Associate the new stabilizers to their respective syndrome circuits
    #   C.1) - Recover the syndrome circuits by name
    xxxx_syndrome_circuit = next(
        syndrome_circuit
        for syndrome_circuit in block.syndrome_circuits
        if syndrome_circuit.name == "xxxx"
    )
    zzzz_syndrome_circuit = next(
        syndrome_circuit
        for syndrome_circuit in block.syndrome_circuits
        if syndrome_circuit.name == "zzzz"
    )
    top_syndrome_circuit = next(
        syndrome_circuit
        for syndrome_circuit in block.syndrome_circuits
        if syndrome_circuit.name == f"top-{stab_top_bottom.lower()}"
    )
    bottom_syndrome_circuit = next(
        syndrome_circuit
        for syndrome_circuit in block.syndrome_circuits
        if syndrome_circuit.name == f"bottom-{stab_top_bottom.lower()}"
    )
    left_syndrome_circuit = next(
        syndrome_circuit
        for syndrome_circuit in block.syndrome_circuits
        if syndrome_circuit.name == f"left-{stab_left_right.lower()}"
    )
    right_syndrome_circuit = next(
        syndrome_circuit
        for syndrome_circuit in block.syndrome_circuits
        if syndrome_circuit.name == f"right-{stab_left_right.lower()}"
    )
    # The syndrome circuits can be reused, they are not modified
    new_syndrome_circuits = block.syndrome_circuits

    #   C.2) - Update the stabilizer to circuit mapping
    new_stab_to_circ = (
        {
            stab.uuid: (
                xxxx_syndrome_circuit.uuid
                if new_upleft_4_body_stabs[0].pauli == "XXXX"
                else zzzz_syndrome_circuit.uuid
            )
            for stab in new_upleft_4_body_stabs
        }
        | {
            stab.uuid: (
                zzzz_syndrome_circuit.uuid
                if new_rest_4_body_stabs[0].pauli == "ZZZZ"
                else xxxx_syndrome_circuit.uuid
            )
            for stab in new_rest_4_body_stabs
        }
        | {stab.uuid: top_syndrome_circuit.uuid for stab in stabs_top}
        | {stab.uuid: bottom_syndrome_circuit.uuid for stab in stabs_bottom}
        | {stab.uuid: left_syndrome_circuit.uuid for stab in stabs_left}
        | {stab.uuid: right_syndrome_circuit.uuid for stab in stabs_right}
    )
    new_stab_to_circ.update(
        {
            k: v
            for (k, v) in block.stabilizer_to_circuit.items()
            if k not in stabs_uuid_to_remove
        }
    )

    if block.boundary_type(direction) == "X":
        update_type = "Z"
        log_ops_untouched = old_x_logicals
        log_ops_to_modify = old_z_logicals
    else:
        update_type = "X"
        log_ops_untouched = old_z_logicals
        log_ops_to_modify = old_x_logicals

    # Detect if qubits involved are not part of the boundary anymore
    qubits_involved = list(log_ops_to_modify[0].data_qubits)
    qubit_to_extend = next(
        q for q in block.boundary_qubits(direction) if q in qubits_involved
    )
    match direction:
        case Direction.LEFT:
            new_qubits = qubits_involved + [
                (qubit_to_extend[0] - i, qubit_to_extend[1], qubit_to_extend[2])
                for i in range(1, length + 1)
            ]
        case Direction.RIGHT:
            new_qubits = qubits_involved + [
                (qubit_to_extend[0] + i, qubit_to_extend[1], qubit_to_extend[2])
                for i in range(1, length + 1)
            ]
        case Direction.TOP:
            new_qubits = qubits_involved + [
                (qubit_to_extend[0], qubit_to_extend[1] - i, qubit_to_extend[2])
                for i in range(1, length + 1)
            ]
        case Direction.BOTTOM:
            new_qubits = qubits_involved + [
                (qubit_to_extend[0], qubit_to_extend[1] + i, qubit_to_extend[2])
                for i in range(1, length + 1)
            ]
    new_logical = PauliOperator(
        pauli=update_type * len(new_qubits),
        data_qubits=new_qubits,
    )
    # D) - Create the new logical operators and keep track of the evolution
    if update_type == "X":
        new_log_x = (new_logical,)
        new_log_z = log_ops_untouched
        interpretation_step.logical_x_evolution.update(
            {new_logical.uuid: (log_ops_to_modify[0].uuid,)}
        )
        # Inherit the updates from the previous operator
        # (new_logical is always a new operator)
        interpretation_step.update_logical_operator_updates_MUT(
            "X", new_logical.uuid, (), True
        )
    else:
        new_log_x = log_ops_untouched
        new_log_z = (new_logical,)
        interpretation_step.logical_z_evolution.update(
            {new_logical.uuid: (log_ops_to_modify[0].uuid,)}
        )
        # Inherit the updates from the previous operator
        # (new_logical is always a new operator)
        interpretation_step.update_logical_operator_updates_MUT(
            "Z", new_logical.uuid, (), True
        )

    new_block = RotatedSurfaceCode(
        unique_label=block.unique_label,
        stabilizers=new_stabilizers,
        logical_x_operators=tuple(new_log_x),
        logical_z_operators=tuple(new_log_z),
        syndrome_circuits=new_syndrome_circuits,
        stabilizer_to_circuit=new_stab_to_circ,
        skip_validation=not debug_mode,
    )
    # Update the block history
    interpretation_step.update_block_history_and_evolution_MUT((new_block,), (block,))

    return interpretation_step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations
from typing import TypeVar

import numpy as np

from .pauli_array import PauliArray
from .pauli_computation import g_npfunc


# Dynamically bound type variable for PauliArray and its subclasses.
T = TypeVar("T", bound=PauliArray)


def ndarray_rowsum(array: np.ndarray, h: int, i: int) -> np.ndarray:
    """
    The rowsum function as described in Aaronson's paper for np.ndarray.
    Reference: https://arxiv.org/abs/quant-ph/0406196

    Parameters
    ----------
    array : np.ndarray
        The array representation of the PauliArray to be modified.
    h : int
        The row-index of the pauli string that will be modified.
    i : int
        The row-index of the pauli string that will be used.

    Returns
    -------
    np.ndarray
        The rowsum'ed array.

    Raises
    ------
    ValueError
        If the rowsum value is odd.
    """
    nqubits = array.shape[1] // 2

    g_array = g_npfunc(
        array[i, :nqubits],
        array[i, nqubits : 2 * nqubits],
        array[h, :nqubits],
        array[h, nqubits : 2 * nqubits],
    )

    sum_g = np.sum(g_array)

    lc_rowsum = 2 * array[h, 2 * nqubits] + 2 * array[i, 2 * nqubits] + sum_g

    if lc_rowsum % 4 == 2:
        array[h, 2 * nqubits] = 1
    elif lc_rowsum % 4 == 0:
        array[h, 2 * nqubits] = 0
    else:
        raise ValueError("rowsum cannot be odd!")

    array[h, :-1] = array[i, :-1] ^ array[h, :-1]

    return array


def rowsum(pauli_array: T, h: int, i: int) -> T:
    """
    The rowsum function as described in Aaronson's paper.
    Reference: https://arxiv.org/abs/quant-ph/0406196

    Parameters
    ----------
    pauli_array : PauliArray
        The PauliArray object to be modified.
    h : int
        The row-index of the pauli string that will be modified.
    i : int
        The row-index of the pauli string that will be used.

    Returns
    -------
    PauliArray
        The rowsum'ed PauliArray.
    """
    pauli_array.array = ndarray_rowsum(pauli_array.array, h, i)
    return pauli_array



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import stim


# pylint: disable=no-member


def noise_annotated_stim_circuit(
    stim_circ: stim.Circuit,
    before_measure_flip_probability: float = 0,
    after_clifford_depolarization: float = 0,
    after_reset_flip_probability: float = 0,
) -> stim.Circuit:
    """
    This function takes as input a pure (sans noise) stim
    circuit, and outputs a the circuit with the desired noise model

    Parameters
    ----------
    stim_circ : stim.Circuit
        The input noiseless stim circuit
    before_measure_flip_probability: float, optional
        X_ERROR probability before a measurement. Default set to
        0 will add no measurement errors
    after_clifford_depolarization: float, optional
        applies DEPOLARIZING_ERROR1 and DEPOLARIZING_ERROR2
        after each single and two qubit clifford gate in the circuit.
        Default set to 0 will add no depolarization errors
    after_reset_flip_probability: float, optional
        Apply an X_ERROR with this probability after a reset gate.
        Default set to 0 will add no reset errors

    Returns
    -------
    stim.Circuit
        stim circuit annotated with the input noise model
    """

    stim_one_qubit_ops = ["H", "X", "Y", "Z", "I"]
    stim_two_qubit_ops = ["CX", "CY", "CZ", "SWAP"]

    def return_annotated_operation(op: stim.CircuitInstruction):
        """
        Append/Prepend an appropriate stim annotation to the
        corresponding stim operation based on how the
        converter has been configured.
        For e.g.
            This operation appends X_ERROR annotation
            before each measurement round, if the
            corresponding error is turned on.
        """
        op_name = op.name
        targets = op.targets_copy()

        annotated_ops_list = [
            {"name": op_name, "targets": targets, "gate_args": op.gate_args_copy()}
        ]

        if op_name == "M" and before_measure_flip_probability > 0:
            annotation = [
                {
                    "name": "X_ERROR",
                    "targets": targets,
                    "gate_args": [before_measure_flip_probability],
                }
            ]
            annotated_ops_list = annotation + annotated_ops_list
        if op_name in stim_one_qubit_ops and after_clifford_depolarization > 0:
            annotation = [
                {
                    "name": "DEPOLARIZE1",
                    "targets": targets,
                    "gate_args": [after_clifford_depolarization],
                }
            ]
            annotated_ops_list = annotated_ops_list + annotation
        if op_name in stim_two_qubit_ops and after_clifford_depolarization > 0:
            annotation = [
                {
                    "name": "DEPOLARIZE2",
                    "targets": targets,
                    "gate_args": [after_clifford_depolarization],
                }
            ]
            annotated_ops_list = annotated_ops_list + annotation
        if op_name == "R" and after_reset_flip_probability > 0:
            annotation = [
                {
                    "name": "X_ERROR",
                    "targets": targets,
                    "gate_args": [after_reset_flip_probability],
                }
            ]
            annotated_ops_list = annotated_ops_list + annotation

        annotated_stim_ops_list = [
            stim.CircuitInstruction(
                name=args_dict["name"],
                targets=args_dict["targets"],
                gate_args=args_dict["gate_args"],
            )
            for args_dict in annotated_ops_list
        ]

        return annotated_stim_ops_list

    annotated_stim_circuit = stim.Circuit()
    for op in stim_circ:
        stim_annotated_op_list = return_annotated_operation(op)
        for annotated_op in stim_annotated_op_list:
            annotated_stim_circuit.append(annotated_op)

    return annotated_stim_circuit



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from functools import reduce
from itertools import product
import numpy as np

from ..eka import LogicalState
from ..eka.utilities import SignedPauliOp, is_tableau_valid, StabArray
from ..cliffordsim import Engine


def get_all_cliffordsim_registers_with_random_flags(
    cat_engine: Engine,
) -> dict[str, tuple[int, bool | None]]:
    """
    Get all the classical registers in the given cliffordsim engine.

    Parameters
    ----------
    cat_engine : Engine
        The cliffordsim engine that contains the classical registers.

    Returns
    -------
    dict[str, tuple[int, bool | None]]
        A dictionary where the keys are the classical register names and the values are
        tuples containing the value of the register and a flag indicating whether the
        register is a result of a random measurement (True) or not (False). If the
        register is not a result of a measurement, the flag is set to None.
    """
    # Get all the classical registers in the cat_engine
    # Set their random flags to None
    res_is_random_dict = {
        bit_id: (bit_value, None)
        for reg in cat_engine.registry.values()
        for bit_id, bit_value in reg.id_bit_reg.items()
    }

    # Look up whether the bit comes as a result of a random measurement
    # Flatten the measurement results to a dict of tuples (m_result, is_random)
    cat_mres_dict = cat_engine.data_store.measurements
    measurement_results = {}
    for time_step in cat_mres_dict["time_step"]:
        bit_id = list(cat_mres_dict[str(time_step)].keys())[0]
        meas_result = cat_mres_dict[str(time_step)][bit_id]["measurement_result"]
        meas_is_random = cat_mres_dict[str(time_step)][bit_id]["is_random"]
        measurement_results[bit_id] = (meas_result, meas_is_random)

    # Update the res_is_random_dict with the measurement results and return it
    res_is_random_dict.update(measurement_results)
    return res_is_random_dict


def get_parity_from_cbits(cat_engine: Engine, cbits: tuple[str | int, ...]) -> int:
    """
    Get the parity of the cbits in the given list.

    Parameters
    ----------
    cat_engine : Engine
        The cliffordsim engine that contains the runtime results.
    cbits : tuple[str | int, ...]
        The list of cbits to check. The cbits can be either strings or integers.
        The integers are treated as constant values (0 or 1) that can flip
        the expected parity. The strings are the labels of the classical channels
        whose values are to be retrieved at runtime. The register name is the first part
        of the label, e.g. c_(0_0)_0 -> c.

    Returns
    -------
    int
        The parity of the cbits. The parity is calculated by XORing the values
        of all of the cbits.
    """
    # Get all the classical registers in the cat_engine and their values
    cliffordsim_classical_registers = get_all_cliffordsim_registers_with_random_flags(
        cat_engine
    )

    # Separate the int cbits from the c_reg values
    # cbit_int_values : Constant values (0 or 1) that can flip the expected parity
    # cbit_labels : Labels of the classical channels whose values are to be
    # retrieved at runtime
    cbit_int_values = [cbit for cbit in cbits if isinstance(cbit, int)]
    cbit_labels = [cbit for cbit in cbits if isinstance(cbit, str)]

    cbit_runtime_values = []
    for cbit_label in cbit_labels:
        # Append the value of the cbit to the list
        cbit_runtime_values += [cliffordsim_classical_registers[cbit_label][0]]

    # Evaluate change in parity
    parity = reduce(
        lambda x, y: x ^ y,
        cbit_int_values + cbit_runtime_values,
        0,
    )

    if parity not in (0, 1):
        raise ValueError(
            "The parity of the cbits is not valid. The parity should be either 0 or 1."
        )

    return parity


def logical_states_to_check(n_logical_qubits: int) -> list[LogicalState]:
    """
    Returns the logical states to check for a given number of logical qubits. These
    logical states are constructed in such a way that if these are transformed
    correctly, then all the logical states should be transformed correctly for a
    Clifford gate operation. The selection is done such that the effect of the gate
    on each input logical operator is isolated.

    For example, to verify the effect on the input logical operators Z1, X1
    (logical qubit 1) for a 3-logical-qubit system, the effect should be captured by
    checking the output of the logical states:
    {+Z0, +Z1, +Z2}, {+Z0, +X1, +Z2}, {+X0, +X1, +X2} and {+X0, +Z1, +X2}.
    We can see that the action of the gate on the logical qubit 1 is isolated by keeping
    the logical qubits 0 and 2 in the same state for the first two and the last two
    states while changing the state of the logical qubit 1.

    NOTE: The above has not been mathematically proven, but it is a reasonable
    assumption that hasn't been disproven yet.

    Parameters
    ----------
    n_logical_qubits : int
        The number of logical qubits.

    Returns
    -------
    list[LogicalState]
        The list of logical states to check.
    """

    if n_logical_qubits == 1:
        # Return the |0> and |+> states
        return [LogicalState(["+Z0"]), LogicalState(["+X0"])]
    if n_logical_qubits == 2:
        # Return the |00>, |++>, |0+>, and |+0> states
        return [
            LogicalState(["+Z0", "+Z1"]),
            LogicalState(["+X0", "+X1"]),
            LogicalState(["+Z0", "+X1"]),
            LogicalState(["+X0", "+Z1"]),
        ]
    return (
        # |00...0> state
        [LogicalState([f"+Z{j}" for j in range(n_logical_qubits)])]
        # |++...+> state
        + [LogicalState([f"+X{j}" for j in range(n_logical_qubits)])]
        # |+0...0>, |0+0...0>, ..., |00...0+> states
        + [
            LogicalState(
                [f"+Z{j}" if i != j else f"+X{j}" for j in range(n_logical_qubits)]
            )
            for i in range(n_logical_qubits)
        ]
        # |0+...+>, |+0+...+>, ..., |++...0> states
        + [
            LogicalState(
                [f"+X{j}" if i != j else f"+Z{j}" for j in range(n_logical_qubits)]
            )
            for i in range(n_logical_qubits)
        ]
    )


def all_possible_pauli_strings(n_qubits: int) -> list[str]:
    """
    Returns all possible sparse Pauli strings for a given number of qubits.
    The sparse Pauli strings are generated by iterating over all possible combinations
    of X and Z operators for each qubit. The list size scales exponentially with the
    number of qubits (as 2*4**n_qubits), so it is important to consider the performance
    implications when working with large number of (logical) qubits.

    Parameters
    ----------
    n_qubits : int
        The number of qubits.
    Returns
    -------
    list[str]
        The list of all possible sparse Pauli strings for the given number of qubits.
    """
    if n_qubits < 1 or not isinstance(n_qubits, int):
        raise ValueError("The number of qubits must be at least 1.")

    paulis = ["_", "X", "Z", "Y"]
    signs = ["+", "-"]
    return [
        f"{sign}{''.join(p)}"
        for sign, p in product(signs, product(paulis, repeat=n_qubits))
        if any(c != "_" for c in p)  # skip all-identity
    ]  # Exclude the empty string (all I's)


# pylint:disable=anomalous-backslash-in-string, line-too-long
def logical_state_transformations_to_check(
    x_operators_sparse_pauli_map: list[str], z_operators_sparse_pauli_map: list[str]
) -> list[tuple[LogicalState, tuple[LogicalState]]]:
    """Returns the logical state transformations to check for a logical operation. The
    input is the sparse Pauli strings for the individual X and Z operators. The
    output is a list of tuples where each tuple contains the input and expected output
    logical states for the logical operation.

    Example:
    For a CNOT gate from qubit 0 to qubit 1 we know that:

    - X0 -> X0X1
    - X1 -> X1
    - Z0 -> Z0
    - Z1 -> Z1Z0

    Thus, we can find the logical state transformations for the CNOT gate by calling
    this function with the following inputs:

    - x_operators_sparse_pauli_map = ["X0X1", "X1"]
    - z_operators_sparse_pauli_map = ["Z0", "Z1Z0"]

    This will return the list of logical state transformations to check for the CNOT
    gate which will be a list containing the following tuples:

    - (LogicalState(('+Z0', '+Z1')), (LogicalState(('+Z0', '+Z0Z1')),)) :math:`\ket{00} -> \ket{00}`
    - (LogicalState(('+X0', '+X1')), (LogicalState(('+X0X1', '+X1')),)) :math:`\ket{++}-> \ket{++}`
    - (LogicalState(('+Z0', '+X1')), (LogicalState(('+Z0', '+X1')),)) :math:`\ket{0+} -> \ket{0+}`
    - (LogicalState(('+X0', '+Z1')), (LogicalState(('+X0X1', '+Z0Z1')),)) :math:`\ket{+0} ->` Bell pair

    The format of the tuples is such that they can be used as input for validator logical
    state transformation checks.

    Parameters
    ----------
    x_operators_sparse_pauli_map : list[str]
        The list of sparse Pauli strings describing how each X operator is transformed.
        No sign is needed and the order matches the transformation of the logical
        operators X0, X1, ...
    z_operators_sparse_pauli_map : list[str]
        The list of sparse Pauli strings describing how each Z operator is transformed.
        No sign is needed and the order matches the transformation of the logical
        operators Z0, Z1, ...

    Returns
    -------
    list[tuple[LogicalState, tuple[LogicalState]]]
        The list of logical state transformations to check.
    """

    if len(z_operators_sparse_pauli_map) != len(x_operators_sparse_pauli_map):
        raise ValueError(
            "The number of X and Z operators should be the same for the logical state "
            "transformations."
        )

    # Get the number of logical qubits
    n_logical_qubits = len(x_operators_sparse_pauli_map)

    # Convert the sparse Pauli strings to SignedPauliOp objects and store them in an
    # np.array for easy access
    x_signed_pauli_op_map = [
        SignedPauliOp.from_sparse_string("+" + x, n_logical_qubits)
        for x in x_operators_sparse_pauli_map
    ]

    z_signed_pauliop_map = [
        SignedPauliOp.from_sparse_string("+" + z, n_logical_qubits)
        for z in z_operators_sparse_pauli_map
    ]

    # For the LogicalStates to be valid, the transformed operators should generate
    # a valid tableau when stacked, similarly to how the X_i and Z_i operators generate
    # a valid tableau when stacked.
    tableau_to_check = np.vstack(
        [x.array for x in x_signed_pauli_op_map]
        + [z.array for z in z_signed_pauliop_map]
    )
    if not is_tableau_valid(tableau_to_check):
        raise ValueError(
            "The transformed X and Z operators do not generate a valid tableau when "
            "stacked. Check the input sparse Pauli string maps."
        )

    x_transform_stabarray = StabArray.from_signed_pauli_ops(x_signed_pauli_op_map)
    z_transform_stabarray = StabArray.from_signed_pauli_ops(z_signed_pauliop_map)

    # Get the input logical states to check
    input_states = logical_states_to_check(n_logical_qubits)

    def _multiply_signed_pauli_ops(
        signed_pauli_ops: list[SignedPauliOp],
        n_logical_qubits: int = n_logical_qubits,
        negative: bool = False,
    ) -> SignedPauliOp:
        """Multiply all of the Pauli operators together and return the result as a
        SignedPauliOp object. Need to specify the number of logical qubits and whether
        the starting sign is negative.
        """
        return reduce(
            lambda x, y: x * y,
            signed_pauli_ops,
            SignedPauliOp.identity(n_logical_qubits, negative=negative),
        )

    # Get the output logical states to check
    output_states = [
        LogicalState(
            # The logical state needs to be defined as a list of sparse Pauli
            # strings which will be found by taking the product of the transformed
            # X and Z operators for each input operator
            [
                # Take the product of the transformed X and Z operators that
                # are present in the input state
                # For example, for the CNOT gate and the input state |00>:
                # we have the input operators +Z0, +Z1
                # which are transformed to +Z0, +Z0Z1 respectively
                _multiply_signed_pauli_ops(
                    # Include all transformed X, Z operators that are present
                    # in the input state (.x==1 or .z==1)
                    x_transform_stabarray[input_pauli_op.x == 1]
                    + z_transform_stabarray[input_pauli_op.z == 1],
                    negative=input_pauli_op.sign == 1,
                ).as_sparse_string()
                # cast to sparse string such that it can be used to create a
                # LogicalState object
                for input_pauli_op in input_state.stabarray
            ]
        )
        for input_state in input_states
    ]

    # Because the output states need to be tuples of LogicalState objects, we need to
    # convert the output states to a tuple of LogicalState objects
    output_states = [(output_state,) for output_state in output_states]

    # Return the input and output logical states as a list of tuples for validator checks
    return list(zip(input_states, output_states, strict=True))



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from importlib.metadata import entry_points

from loom.eka import Circuit, Eka
from loom.eka.operations import Operation, BaseOperation, CodeOperation

from .applicator import BaseApplicator, CodeApplicator
from .interpretation_step import InterpretationStep


def load_plugin_options():
    """
    Load plugin options from entry points defined in the 'loom.selectors' group.
    If plugins are installed, they are detected by this method. This allows the user to
    use `interpret_eka` with plugin specific applicators.
    """
    plugin_options = {}
    for ep in entry_points().select(group="loom.selectors"):
        data_provider = ep.load()
        plugin_options.update(data_provider)
    return plugin_options


def interpret_operation(
    eka: Eka,
    step: InterpretationStep,
    op: Operation,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    This function interprets the given operation and returns a new InterpretationStep
    which contains all modifications due to the operation. The function itself finds the
    right applicator for the given operation and calls it. The actual implementation of
    the operation is done in the applicator.

    Parameters
    ----------
    eka : Eka
        The Eka object for which the operation should be interpreted. The Eka is
        passed on to the applicator. This allows the applicator to access the lattice
        and the syndrome circuits
    step : InterpretationStep
        Current InterpretationStep to which the modifications due to the operation
        should be applied
    op : :class:`loom.eka.operations.base_operation.Operation`
        Operation to be interpreted
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Currently, the effects of debug mode are:
        - Disabling the commutation validation of Block

    Returns
    -------
    InterpretationStep
        New InterpretationStep containing all modifications due to the operation
    """
    match op:
        case BaseOperation():
            # BaseOperations are implemented at the level of the BaseApplicator
            return BaseApplicator(eka).apply(step, op, same_timeslice, debug_mode)
        case CodeOperation():
            # Get the class name of the blocks
            block_class_names = list(
                set(
                    step.get_block(block_label).__class__.__name__
                    for block_label in op._inputs  # pylint: disable=protected-access
                )
            )
            # For CodeOperations, check that the code types of all input blocks are the
            # same and call the appropriate applicator
            if len(block_class_names) > 1:
                raise NotImplementedError(
                    "Operations including blocks of different classes are not "
                    f"supported at the moment. Block classes: {block_class_names}"
                )
            if block_class_names[0] == "Block":
                return CodeApplicator(eka).apply(step, op, same_timeslice, debug_mode)
            plugin_options = load_plugin_options()
            if plugin_options:
                # If there are plugins available, we can try to load the appropriate
                # applicator for the block class name
                for plugin_value in plugin_options.values():
                    if plugin_value["block_class_name"] == block_class_names[0]:
                        return plugin_value["applicator"](eka).apply(
                            step, op, same_timeslice, debug_mode
                        )
            raise NotImplementedError(
                f"The Block type '{block_class_names[0]}' "
                "is not supported at the moment"
            )

    if step.composite_operation_session_stack:
        raise ValueError(
            "Please ensure that all composite operation sessions have been ended. "
            f"Found live sessions: {step.composite_operation_session_stack}"
        )


def cleanup_final_step(final_step: InterpretationStep) -> InterpretationStep:
    """
    Clean up the final interpretation step before it is returned to the user.

    Parameters
    ----------
    final_step : InterpretationStep
        Final interpretation step which should be cleaned up before returning it to the
        user

    Returns
    -------
    InterpretationStep
        Cleaned up interpretation step
    """
    cleaned_final_step = final_step
    # Remove all channels from the channel_dict that are not part of the circuit
    # This can only be done if the circuit is not None (which is true for all realistic
    # use cases). For empty circuits, the dict is set to an empty dict.

    # circuit_seq is a tuple of tuples of composite circuits, we need to align the empty
    # tuples to make sure that the final circuit has the correct duration and that
    # parallel operations are indeed in parallel
    circuit_seq = final_step.intermediate_circuit_sequence
    full_circuit = ()
    for timeslice in circuit_seq:
        timespan = max(composite_circuit.duration for composite_circuit in timeslice)
        # Create a template circuit
        template_circ = (
            tuple(composite_circuit for composite_circuit in timeslice),
        ) + ((),) * (timespan - 1)
        full_circuit += template_circ

    cleaned_final_step.final_circuit = Circuit(
        name="Final circuit",
        circuit=full_circuit,
        channels=sorted(  # Ensure channels are sorted
            list(
                set(
                    channel
                    for timeslice in full_circuit
                    for circuit in timeslice
                    for channel in circuit.channels
                )
            ),
            key=lambda ch: ch.label,
        ),
    )

    cleaned_final_step.channel_dict = (
        {
            ch.id: ch
            for ch in cleaned_final_step.channel_dict.values()
            if ch in cleaned_final_step.final_circuit.channels
        }
        if cleaned_final_step.final_circuit is not None
        else {}
    )
    return cleaned_final_step


def interpret_eka(eka: Eka, debug_mode: bool = True) -> InterpretationStep:
    """
    Interpret the Eka object and return the final InterpretationStep. The function
    iterates over all operations in the Eka and applies them to the current
    InterpretationStep. The function also handles the case where multiple operations
    are applied in parallel during the same timestep.

    Parameters
    ----------
    eka : Eka
        Eka object describing the operations to perform, as well as the initial state
         of the blocks.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block

    Returns
    -------
    InterpretationStep
        Final InterpretationStep containing the interpreted operations and the final
        circuit.
    """
    # Initialize the interpretation step by passing the initial blocks as the first
    # element of the block history
    step = InterpretationStep.create(initial_blocks=eka.blocks)
    for timestep in eka.operations:
        # Reset the same_timeslice flag for each new timestep
        same_timeslice = False
        for op in timestep:
            step = interpret_operation(eka, step, op, same_timeslice, debug_mode)
            # Set the same_timeslice flag to True after the first operation in the
            # timestep has been interpreted to indicate that the following operations
            # are part of the same timestep
            same_timeslice = True

    # Return a cleaned up version of the final step
    step = cleanup_final_step(step)
    # Freeze the InterpretationStep to prevent further modifications
    step.is_frozen = True
    return step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from dataclasses import asdict
from pydantic.dataclasses import dataclass

from loom.eka.utilities import dataclass_config

from ..stabilizer import Stabilizer
from ..pauli_operator import PauliOperator


@dataclass(config=dataclass_config)
class Operation:
    """
    Parent class for all operations in the EKA.
    """

    def asdict(self):
        """Serialize the operation to a dictionary."""
        # Use the dataclass's asdict method to convert the instance to a dictionary
        class_dict = asdict(self)
        # Add the class name to the dictionary for deserialization
        class_dict["__name__"] = self.__class__.__name__
        return class_dict

    @classmethod
    def fromdict(cls, data_dict: dict):
        """Deserialize the operation from a dictionary."""
        # The class name should be included in the dictionary
        # (not needed for the dataclass initialization, only for deserialization)
        if not (cls_name := data_dict.pop("__name__")):
            raise ValueError(
                "No Operation name found in the data, the Operation cannot be loaded."
            )
        # If the right class is given, we can just instantiate it
        if cls_name == cls.__name__:
            return cls(**data_dict)

        # For abstract classes, we need to look through the subclasses
        operation_class = next(
            (
                subsubclass
                for subclass in cls.__subclasses__()  # Base, Code or Logical (abstract)
                for subsubclass in subclass.__subclasses__()  # Subsequent subclasses
                if subsubclass.__name__ == cls_name
            ),
            None,
        )
        if operation_class is None:
            raise ValueError(
                f"Operation {cls_name} was not found in the Operation subclasses."
            )

        return operation_class(**data_dict)


@dataclass(config=dataclass_config)
class BaseOperation(Operation):
    """
    Base class for all operations in the EKA.

    NOTE: Base Operations are Operations that can be generalized to any setup of the
    EKA. They are not specific to a particular code or setup. This could range from
    single qubit gates on specific qubits to measuring specific stabilizers of the code.
    """


@dataclass(config=dataclass_config)
class MeasureStabilizerSyndrome(BaseOperation):
    """
    Measure the syndrome of a single stabilizer.

    Parameters
    ----------

    stabilizer : Stabilizer
        Stabilizer to measure.
    """

    stabilizer: Stabilizer


@dataclass(config=dataclass_config)
class MeasureObservable(BaseOperation):
    """
    Measure an observable.

    Parameters
    ----------

    observable : PauliOperator
        Observable to measure.
    """

    observable: PauliOperator



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations
from uuid import uuid4

from pydantic.dataclasses import dataclass
from pydantic import field_validator, Field
import numpy as np

from .utilities.validation_tools import (
    nr_of_qubits_error,
    distinct_error,
    dataclass_config,
    ensure_tuple,
    coordinate_length_error,
    pauli_error,
)
from .utilities.pauli_format_conversion import paulichar_to_xz, paulixz_to_char_npfunc
from .utilities.pauli_binary_vector_rep import SignedPauliOp


@dataclass(config=dataclass_config)
class PauliOperator:
    """
    A PauliOperator is defined by a pauli string, and a set of data qubits.

    Parameters
    ----------
    pauli: str
        The Pauli string that defines this operator.
    data_qubits: tuple[tuple[int, ...], ...]
        Qubits involved in the operator. They are referred to by their coordinates
        in the lattice.
    uuid : str
        Unique identifier of the operator. This is automatically set to a random UUID.
    """

    pauli: str
    data_qubits: tuple[tuple[int, ...], ...]
    uuid: str = Field(default_factory=lambda: str(uuid4()), validate_default=True)

    # Validation functions
    _validate_pauli = field_validator("pauli")(pauli_error)
    _validate_qubits_list = field_validator("data_qubits", mode="before")(ensure_tuple)
    _validate_number_qubits = field_validator("data_qubits")(nr_of_qubits_error)
    _validate_distinct_qubits = field_validator("data_qubits")(distinct_error)
    _validate_coordinate_lengths_qubits = field_validator("data_qubits", mode="before")(
        coordinate_length_error
    )

    # Magic methods
    def __str__(self) -> str:
        pauli_ops = [
            f"{p}_{idx}" for p, idx in zip(self.pauli, self.data_qubits, strict=True)
        ]
        return " ".join(pauli_ops)

    def __repr__(self) -> str:
        # use the __str__ method to represent the PauliOperator but add the class name
        return f"{self.__class__.__name__}({self})"

    def __eq__(self, other: PauliOperator) -> bool:
        """
        Ignore the uuid in the equality check.
        """
        if not isinstance(other, PauliOperator):
            return NotImplemented
        return dict(zip(self.data_qubits, self.pauli, strict=True)) == dict(
            zip(other.data_qubits, other.pauli, strict=True)
        )

    # Properties

    @property
    def weight(self) -> int:
        """Number of qubits involved in the operator."""
        return len(self.data_qubits)

    @property
    def pauli_type(self) -> str:
        """Type of the Pauli operator: 'X', 'Y', or 'Z'."""
        unique_paulis = set(self.pauli)
        if len(unique_paulis) != 1:
            raise ValueError(
                "PauliOperator must consist of a single type of Pauli operator "
                f"to determine its type, got {unique_paulis} instead."
            )
        return unique_paulis.pop()

    # Methods
    def as_signed_pauli_op(
        self, all_qubits: tuple[tuple[int, ...], ...]
    ) -> SignedPauliOp:
        """
        Get the SignedPauliOp representation of the PauliOperator.

        Parameters
        ----------
        all_qubits: tuple[tuple[int, ...], ...]
            All qubits coordinates in the system.

        Returns
        -------
        SignedPauliOp
            The SignedPauliOp representation of the PauliOperator.

        Raises
        ------
        ValueError
            If the number of qubits in the system is less than the number of qubits in
            the operator.
        """
        all_qubits = tuple(map(tuple, all_qubits))
        if len(all_qubits) < len(self.data_qubits):
            raise ValueError(
                f"Number of qubits in the operator {len(self.data_qubits)} exceeds the "
                f"total number of qubits in the system {len(all_qubits)}."
            )

        # Get the x and z values for each qubit in the operator
        x_values, z_values = tuple(
            zip(*[paulichar_to_xz(p) for p in self.pauli], strict=True)
        )
        # Get the sign of the operator
        sign = 0
        # Cast the indexed dqubits to a numpy array for indexing
        all_qubits_map = {q: i for i, q in enumerate(all_qubits)}
        idx_dqubits = np.array(
            [all_qubits_map[q] for q in self.data_qubits if q in all_qubits_map]
        )

        # Initialize the operator row
        op_row = np.zeros(2 * len(all_qubits) + 1, dtype=SignedPauliOp.DTYPE)
        # Fill the operator row with the x, z values and the sign
        op_row[idx_dqubits] = x_values
        op_row[len(all_qubits) + idx_dqubits] = z_values
        op_row[-1] = sign

        # Use the op_row to create a SignedPauliOp object that is already validated
        return SignedPauliOp(op_row, validated=True)

    @staticmethod
    def from_signed_pauli_op(
        signed_pauli_op: SignedPauliOp, index_to_qubit_map: dict[int, tuple[int, ...]]
    ) -> PauliOperator:
        """
        Create a PauliOperator from a SignedPauliOp.

        Parameters
        ----------
        signed_pauli_op : SignedPauliOp
            The SignedPauliOp to convert to a PauliOperator.
        index_to_qubit_map : dict[int, tuple[int, ...]]
            A dictionary mapping the indices of the SignedPauliOp to the qubit
            coordinates in the lattice.

        Returns
        -------
        PauliOperator
            The PauliOperator representation of the SignedPauliOp.
        """
        # Find the indexed_dqubits
        indexed_dqubits = np.union1d(
            np.where(signed_pauli_op.x != 0)[0],
            np.where(signed_pauli_op.z != 0)[0],
        )

        # Check if all the indexed dqubits have a corresponding qubit in the map
        missing_indices = set(indexed_dqubits) - set(index_to_qubit_map.keys())
        if missing_indices:
            # Sort and cast them into a sorted list of integers
            missing_indices = sorted(map(int, missing_indices))
            raise ValueError(
                f"Missing qubit coordinates for indices {missing_indices}."
            )

        # Get the pauli string of the indexed dqubits
        pauli_str = "".join(
            paulixz_to_char_npfunc(
                signed_pauli_op.x[indexed_dqubits], signed_pauli_op.z[indexed_dqubits]
            )
        )

        return PauliOperator(
            pauli=pauli_str,
            data_qubits=tuple(index_to_qubit_map[i] for i in indexed_dqubits),
        )

    def commutes_with(self, other_operator: PauliOperator) -> bool:
        """
        Check if the PauliOperator commutes with another PauliOperator.

        Parameters
        ----------
        other_operator : PauliOperator
            The other PauliOperator to check commutation with.

        Returns
        -------
        bool
            True if the two objects commute, False otherwise.
        """
        if not isinstance(other_operator, PauliOperator):
            raise ValueError(
                "Expected PauliOperator object, got " f"{type(other_operator)} instead."
            )
        # Find common qubits
        common_qubits = set(self.data_qubits).intersection(
            set(other_operator.data_qubits)
        )

        # Find for each of the common qubits whether their paulis anti-commute.
        # They anti-commute if their paulis are different.
        anti_commutation_of_common_qubits = [
            self.pauli[self.data_qubits.index(qubit)]
            # the above is the pauli of the self PauliOperator for the qubit
            !=
            # the one below is the pauli of the other PauliOperator for the qubit
            other_operator.pauli[other_operator.data_qubits.index(qubit)]
            for qubit in common_qubits
        ]

        # Return the total commutation of the common qubits
        return not bool(np.sum(anti_commutation_of_common_qubits) % 2)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import numpy as np


def paulis_anti_commute(x1: int, z1: int, x2: int, z2: int) -> int:
    """
    Calculates the anti-commutation value of two pauli operators.

    A value of 0 means that the paulis commute while a value of 1 means that
    they anti-commute.
    """
    return (x1 & z2) ^ (z1 & x2)


def anti_commutes_npfunc(
    x1: np.ndarray, z1: np.ndarray, x2: np.ndarray, z2: np.ndarray
) -> np.ndarray:
    """
    Vectorized anti-commutation function.

    Parameters
    ----------
    x1 : np.ndarray
        The x bits of the first pauli string.
    z1 : np.ndarray
        The z bits of the first pauli string.
    x2 : np.ndarray
        The x bits of the second pauli string.
    z2 : np.ndarray
        The z bits of the second pauli string.

    Returns
    -------
    np.ndarray
        The anti-commutation values.
    """
    return np.frompyfunc(paulis_anti_commute, 4, 1)(x1, z1, x2, z2)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Circuit, PauliOperator, Stabilizer
from loom.eka.operations import Grow
from loom.eka.utilities import Direction
from loom.interpreter import InterpretationStep

from ..code_factory import RepetitionCode


# pylint: disable=duplicate-code, disable=too-many-locals
def grow_consistency_check(
    interpretation_step: InterpretationStep, operation: Grow
) -> RepetitionCode:
    """
    Check the consistency of the grow operation.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        InterpretationStep containing the block to grow.
    operation : Grow
        Grow operation description.

    Returns
    -------
        The RepetitionCode block to grow.

    Raises
    ------
    ValueError
        If the Block is not RepetitionCode.
        If the grown operation direction is not RIGHT or LEFT.
        If growing in the left direction goes beyond the lattice boundary.
    """
    # Extract Block
    block = interpretation_step.get_block(operation.input_block_name)

    # Check valid Block type
    if not isinstance(block, RepetitionCode):
        raise ValueError(
            f"This grow operation is not supported for {type(block)} blocks."
        )

    # Check grow is applied in the correct direction
    if operation.direction not in [Direction.RIGHT, Direction.LEFT]:
        raise ValueError(
            "Repetition code does not support growing in the "
            f"{operation.direction} direction."
        )
    # Check that we are not growing beyond the boundaries of the lattice
    if operation.direction == Direction.LEFT:
        left_boundary_position = block.boundary_qubits(Direction.LEFT)[0]
        if left_boundary_position - operation.length < 0:
            raise ValueError("Cannot grow beyond the boundary of the lattice.")

    return block


def get_new_data_qubits_info(
    block: RepetitionCode, direction: Direction, length: int
) -> list[tuple[int, int]]:
    """Find the qubits to add during the grow operation.

    Parameters
    ----------
    block : RepetitionCode
        Block to grow.
    direction : Direction
        Direction of the grow operation.
    length : int
        Length of the grow operation.

    Returns
    -------
    list[tuple[int, int]]
        List containing the new qubits to add.
    """

    left_boundary_qubit = block.boundary_qubits(Direction.LEFT)

    # Find the new data qubits that need to be added
    match direction:
        case Direction.LEFT:
            new_data_qubits = [
                (-i + left_boundary_qubit[0] - 1, 0) for i in range(length)
            ]
        case Direction.RIGHT:
            right_boundary_qubit = block.boundary_qubits(Direction.RIGHT)
            new_data_qubits = [
                (i + right_boundary_qubit[0] + 1, 0) for i in range(length)
            ]

    return new_data_qubits


def create_grow_circuit(
    interpretation_step: InterpretationStep,
    check_type: str,
    new_data_qubits: list[tuple[int, int]],
    circuit_name: str,
) -> Circuit:
    """
    Generates the circuit required to reset the new data that get added during the
    grow operation.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the block to grow.
    check_type : str
        Type of stabilizers in the code.
    new_data_qubits : list[tuple[int,int]]
        List of new data qubits to reset.
    circuit_name : str
        Name of the circuit.

    Returns
    -------
    Circuit
        Circuit to reset the new data qubits.
    """

    # Select reset state based on stabilizer type
    if check_type == "X":
        reset_state = "0"
    else:
        reset_state = "+"

    # Reset sequence
    reset_sequence = [
        [
            Circuit(
                name=f"reset_{reset_state}",
                channels=interpretation_step.get_channel_MUT(q, "quantum"),
            )
            for q in new_data_qubits
        ]
    ]

    # Circuit for the grow operation
    grow_circuit = Circuit(
        name=circuit_name,
        circuit=reset_sequence,
    )

    return grow_circuit


def find_new_stabilizers(
    block: RepetitionCode,
    check_type: str,
    is_left: bool,
    new_data_qubits: list[tuple[int, int]],
) -> list[Stabilizer]:
    """Find the new stabilizers that will form the grown block.

    Parameters
    ----------
    block : RepetitionCode
        Block to grow.
    check_type : str
        Type of stabilizer to add.
    is_left: bool
        Whether the grow operation is to the left or right.
    new_data_qubits : list[tuple[int,int]]
        New data qubits to add.

    Returns
    -------
    list[Stabilizer]
        List containing the new stabilizers forming the grown block.
    """

    # Define the stabilizers to add
    stabs_to_add = [
        Stabilizer(
            pauli=check_type * 2,
            data_qubits=sorted(
                [(coord - (1 - 2 * is_left), 0), (coord, 0)], key=lambda x: x[0]
            ),
            ancilla_qubits=[(coord - (1 - is_left), 1)],
        )
        for coord, _ in new_data_qubits
    ]
    # Combine the stabilizers to get the new set of stabilizers
    new_stabilizers = list(block.stabilizers) + stabs_to_add

    return new_stabilizers


def get_logical_operator_and_evolution(
    block: RepetitionCode,
    check_type: str,
    new_data_qubits: list[tuple[int, int]],
) -> tuple[list[list[PauliOperator]], list[dict[str, tuple[str, ...]]]]:
    """
    Extract the logical operators and the evolution after the grow operation.

    Parameters
    ----------
    block : RepetitionCode
        Block to grow.
    check_type : str
        Type of stabilizer defining the code.
    new_data_qubits : list[tuple[int,int]]
        New data qubits to add.

    Returns
    -------
    tuple[list[list[PauliOperator]], list[dict[str, tuple[str, ...]]]]
        Tuple containing the new logical operators, the evolution of the logical X
        operators, and the evolution of the logical Z operators.
    """

    complementary_check_type = "X" if check_type == "Z" else "Z"

    # Combine old and new data qubits
    all_qubits = list(block.data_qubits) + new_data_qubits

    # Extract old long and short operators
    old_logs = [block.logical_x_operators[0], block.logical_z_operators[0]]
    old_short_log, old_long_log = sorted(old_logs, key=lambda x: len(x.data_qubits))

    # Define new logical operators
    new_long_log = PauliOperator(
        pauli=complementary_check_type * len(all_qubits), data_qubits=all_qubits
    )

    new_short_log = old_short_log

    # New logicals
    new_logs = [[new_long_log], [new_short_log]]

    # Log evolution
    long_log_evolution = {new_long_log.uuid: (old_long_log.uuid,)}
    short_log_evolution = {}
    log_evolution = [long_log_evolution, short_log_evolution]

    return new_logs, log_evolution


def grow(  # pylint: disable=too-many-locals
    interpretation_step: InterpretationStep,
    operation: Grow,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Grow a Repetition Code chain in the specified direction (left or right).

    The algorithm is the following:

    - A.) DATA QUBITS

        - A.1) Extract new data qubits

    - B.) CIRCUIT

        - B.1) Generate the circuit
        - B.2) Add the circuit to the interpretation step

    - C.) STABILIZERS

        - C.1) Find the new stabilizers that get added
        - C.2) Add to current set of stabilizers

    - D.) LOGICAL OPERATORS

        - D.1) Extract logical operators and updates
        - D.2) Update the logical operator history

    - E.) NEW BLOCK

        - E.1) Create the new block
        - E.2) Update the block history

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the block to grow.
    operation : Grow
        Grow operation description.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block.

    Returns
    -------
    InterpretationStep
        Interpretation step after the grow operation.
    """

    # Check consistency of the grow operation
    block = grow_consistency_check(interpretation_step, operation)

    # Extract stabilizer type
    check_type = block.check_type
    is_left = operation.direction == Direction.LEFT

    # A) DATA QUBITS
    # A.1) Extract new data qubits
    new_data_qubits = get_new_data_qubits_info(
        block, direction=operation.direction, length=operation.length
    )

    # B) CIRCUIT
    # B.1) Generate the circuit
    circuit_name = (
        f"grow {block.unique_label} by {operation.length} to the {operation.direction}"
    )
    grow_circuit = create_grow_circuit(
        interpretation_step, check_type, new_data_qubits, circuit_name
    )

    # B.2) Add the circuit to the interpretation step
    interpretation_step.append_circuit_MUT(grow_circuit, same_timeslice)

    # C) STABILIZERS
    # C.1) Find the new set of stabilizers
    new_stabilizers = find_new_stabilizers(block, check_type, is_left, new_data_qubits)

    # D) LOGICAL OPERATORS
    # D.1) Extract logical operators and updates
    new_logs, log_evolution = get_logical_operator_and_evolution(
        block,
        check_type,
        new_data_qubits,
    )

    # D.2) Update the logical operator history
    ordering = 1 if check_type == "Z" else -1
    new_log_x_ops, new_log_z_ops = new_logs[::ordering]
    logical_x_evolution, logical_z_evolution = log_evolution[::ordering]

    # Update the logical operator evolution
    interpretation_step.logical_x_evolution.update(logical_x_evolution)
    interpretation_step.logical_z_evolution.update(logical_z_evolution)
    # Update the logical operator updates
    for op in new_log_x_ops:
        interpretation_step.update_logical_operator_updates_MUT("X", op.uuid, (), True)
    for op in new_log_z_ops:
        interpretation_step.update_logical_operator_updates_MUT("Z", op.uuid, (), True)

    # E) NEW BLOCK
    # E.1) Create the new block
    new_block = RepetitionCode(
        stabilizers=new_stabilizers,
        logical_x_operators=new_log_x_ops,
        logical_z_operators=new_log_z_ops,
        unique_label=block.unique_label,
        skip_validation=not debug_mode,
    )

    # E.2) Update the block history
    interpretation_step.update_block_history_and_evolution_MUT(
        new_blocks=(new_block,),
        old_blocks=(block,),
    )

    return interpretation_step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from collections import defaultdict

import numpy as np
import networkx as nx
import plotly.graph_objs as go

from loom.eka import Circuit

from .plotting_utils import convert_circuit_to_nx_graph


def hierarchy_layout(graph, root=None, x_spacing=1.0, y_spacing=1.5):
    """
    Pure-Python replacement for graphviz_layout (hierarchical layout).

    Parameters
    ----------
    graph : nx.DiGraph
        Directed acyclic graph to layout.
    root : node or None
        Optional root node. If None, the first node with in-degree 0 is used.
    x_spacing : float
        Horizontal spacing between sibling nodes.
    y_spacing : float
        Vertical spacing between levels.

    Returns
    -------
    dict
        Mapping of node -> (x, y) positions.
    """

    if not nx.is_directed_acyclic_graph(graph):
        graph = nx.DiGraph(graph)

    # Choose a root if none provided
    if root is None:
        roots = [n for n, deg in graph.in_degree() if deg == 0]
        root = roots[0] if roots else list(graph.nodes())[0]

    # BFS/topological sort to determine depth/layer of each node
    layers = defaultdict(list)
    for node in nx.topological_sort(graph):
        preds = list(graph.predecessors(node))
        depth = (
            0 if not preds else 1 + max(graph.nodes[p].get("_depth", 0) for p in preds)
        )
        graph.nodes[node]["_depth"] = depth
        layers[depth].append(node)

    # Assign x, y positions
    pos = {}
    max_width = max(len(nodes) for nodes in layers.values())
    for depth, nodes in layers.items():
        width = len(nodes)
        for i, node in enumerate(nodes):
            x = (i - (width - 1) / 2) * x_spacing * (max_width / width)
            y = -depth * y_spacing
            pos[node] = (x, y)
    return pos


# pylint: disable=too-many-locals
def plot_circuit_tree(
    circuit: Circuit,
    max_layer: int | None = None,
    layer_colors: list[str] | None = None,
    layer_labels: list[str] | None = None,
    num_layers_with_text: int | None = 1,
) -> go.Figure:
    """
    Plot the tree structure of a `Circuit` object.

    Parameters
    ----------
    circuit: Circuit
        Circuit which should be plotted
    max_layer: int | None
        Maximum layer up to which the tree should be plotted.
        If None is provided, all layers are plotted.
    layer_colors: list[str] | None
        Array of colors for the markers in different layers.
    layer_labels: list[str] | None
        Array of labels for the different layers.
        If None is provided, they are labeled by their number.
    num_layers_with_text: int | None
        Number of layers where their name is written on top of the markers

    Returns
    -------
    go.Figure
        Interactive `go.Figure` object for the circuit tree
    """

    # Default colors for the different layers
    if layer_colors is None:
        layer_colors = [
            "#054352",
            "#536070",
            "#857972",
            "#94AD72",
            "#D1E071",
            "#B39F67",
            "#946F52",
            "#D16F4D",
            "#E64040",
            "#990538",
            "#61105B",
            "#3F2882",
            "#005FA8",
        ]

    # Get an nx graph representing the circuit and a list of labels for all nodes
    graph, labels_nodes = convert_circuit_to_nx_graph(circuit)

    # Tree-like layout (pure Python, no Graphviz required)
    positions = hierarchy_layout(graph)

    nodes_x_coords = [x for x, y in positions.values()]
    nodes_y_coords = [y for x, y in positions.values()]

    # Create mapping from node ID to coordinates
    node_positions = {node: (x, y) for node, (x, y) in positions.items()}
    # Create lists of edge coordinates
    edge_x_coords = []
    edge_y_coords = []
    for u, v in graph.edges():
        x0, y0 = node_positions[u]
        x1, y1 = node_positions[v]
        edge_x_coords.append([x0, x1, None])
        edge_y_coords.append([y0, y1, None])

    # Sorted list of unique y-coordinates for layers
    nodes_y_coords_set = sorted({y for _, y in node_positions.values()}, reverse=True)

    # Check whether provided layer labels are valid
    if layer_labels is None:
        layer_labels = [
            f"Layer {layer_idx+1}" for layer_idx in range(len(nodes_y_coords_set))
        ]
    else:
        if len(layer_labels) != len(nodes_y_coords_set):
            raise ValueError(
                "Number of layer labels does not match the "
                f"number of layers. {len(layer_labels)} layer "
                f"labels were provided while there are "
                f"{len(nodes_y_coords_set)} layers."
            )

    fig = go.Figure()

    # This dummy trace is not visible but needed for the correct ordering of layers
    fig.add_trace(
        go.Scatter(
            x=[None],
            y=[None],
            name="Layer 1",
            mode="lines",
            line={"color": "rgb(210,210,210)", "width": 1},
            hoverinfo="none",
            legendgroup="layer_1",
            showlegend=False,
        )
    )

    # Create the condition for two elements close on the y axis
    def is_close_in_y(y1, y2, tol=1e-3):
        """Check if two y-coordinates are close to each other."""
        return np.abs(y1 - y2) < tol

    # Draw lines from a circuit to its subcircuits
    for layer_idx, y_val in enumerate(nodes_y_coords_set):
        if max_layer is not None and layer_idx >= max_layer - 1:
            break
        points_in_this_layer = np.array(
            [
                (xs, ys)
                for xs, ys in zip(edge_x_coords, edge_y_coords, strict=True)
                if is_close_in_y(ys[0], y_val, tol=1e-3)
            ]
        )
        if len(points_in_this_layer) > 0:
            fig.add_trace(
                go.Scatter(
                    x=points_in_this_layer[:, 0].flatten(),
                    y=points_in_this_layer[:, 1].flatten(),
                    name=f"Layer {layer_idx+2}",
                    mode="lines",
                    line={"color": "rgb(210,210,210)", "width": 1},
                    hoverinfo="none",
                    legendgroup=f"layer_{layer_idx+2}",
                    showlegend=False,
                )
            )

    # Plot nodes
    for layer_idx, y_val in enumerate(nodes_y_coords_set):
        if max_layer is not None and layer_idx >= max_layer:
            break
        background_color = layer_colors[layer_idx % len(layer_colors)]
        points_in_this_layer = np.array(
            [
                (x, y, label)
                for x, y, label in zip(
                    nodes_x_coords, nodes_y_coords, labels_nodes, strict=True
                )
                if is_close_in_y(y, y_val, tol=1e-3)
            ]
        )
        if num_layers_with_text is not None and layer_idx < num_layers_with_text:
            mode = "markers+text"
            hoverinfo = "none"
        else:
            mode = "markers"
            hoverinfo = "text"
        fig.add_trace(
            go.Scatter(
                x=points_in_this_layer[:, 0],
                y=points_in_this_layer[:, 1],
                mode=mode,
                name=layer_labels[layer_idx],
                marker={
                    "symbol": "circle-dot",
                    "size": 18,
                    "color": background_color,
                    "line": {"color": "rgb(50,50,50)", "width": 1},
                },
                text=points_in_this_layer[:, 2],
                textposition="top center",
                hoverinfo=hoverinfo,
                opacity=0.8,
                legendgroup=f"layer_{layer_idx+1}",
            )
        )

    fig.update_layout(
        xaxis={
            "showgrid": False,
            "zeroline": False,
            "visible": False,
        },
        yaxis={
            "showgrid": False,
            "visible": False,
        },
    )

    return fig



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from dataclasses import dataclass, field

from .base_operation import Operation, OpType
from .controlled_operation import has_ccontrol


@dataclass
@has_ccontrol
class ResizeOperation(Operation):
    """
    Operations of this type resize the number of qubits in the Engine during runtime.
    """

    operation_type: str = field(default=OpType.RESIZE, init=False)
    target_qubit: int


@dataclass
class AddQubit(ResizeOperation):
    """
    An Operation that adds a qubit to the Engine during runtime. The qubit will be
    initialised in the `|0>` state.
    """

    name: str = field(default="AddQubit", init=False)


@dataclass
class DeleteQubit(ResizeOperation):
    """
    An Operation that deletes a qubit from the Engine during runtime.
    """

    name: str = field(default="DeleteQubit", init=False)
